# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Paul Michali, Cisco Systems, Inc.

import httplib
import os
import tempfile

import mock

from neutron import context
from neutron.openstack.common import uuidutils
from neutron.plugins.common import constants
from neutron.services.vpn.device_drivers import (
    cisco_csr_rest_client as csr_client)
from neutron.services.vpn.device_drivers import cisco_ipsec as ipsec_driver
from neutron.tests import base

_uuid = uuidutils.generate_uuid
FAKE_HOST = 'fake_host'
FAKE_ROUTER_ID = _uuid()
FAKE_VPN_SERVICE = {
    'id': _uuid(),
    'router_id': FAKE_ROUTER_ID,
    'admin_state_up': True,
    'status': constants.PENDING_CREATE,
    'subnet': {'cidr': '10.0.0.0/24'},
    'ipsec_site_connections': [
        {'peer_cidrs': ['20.0.0.0/24',
                        '30.0.0.0/24']},
        {'peer_cidrs': ['40.0.0.0/24',
                        '50.0.0.0/24']}]
}
FIND_CFG_FOR_CSRS = ('neutron.services.vpn.device_drivers.cisco_ipsec.'
                     'find_available_csrs_from_config')


class TestCiscoCsrIPSecConnection(base.BaseTestCase):
    def setUp(self):
        super(TestCiscoCsrIPSecConnection, self).setUp()
        self.conn_info = {
            u'id': '123',
            u'status': constants.PENDING_CREATE,
            u'admin_state_up': True,
            'psk': 'secret',
            'peer_address': '192.168.1.2',
            'peer_cidrs': ['10.1.0.0/24', '10.2.0.0/24'],
            'mtu': 1500,
            'ike_policy': {'auth_algorithm': 'sha1',
                           'encryption_algorithm': 'aes-128',
                           'pfs': 'Group5',
                           'ike_version': 'v1',
                           'lifetime_units': 'seconds',
                           'lifetime_value': 3600},
            'ipsec_policy': {'transform_protocol': 'ah',
                             'encryption_algorithm': 'aes-128',
                             'auth_algorithm': 'sha1',
                             'pfs': 'group5',
                             'lifetime_units': 'seconds',
                             'lifetime_value': 3600},
            'cisco': {'site_conn_id': 'Tunnel0',
                      'ike_policy_id': 222,
                      'ipsec_policy_id': 333,
                      # TODO(pcm) FUTURE use vpnservice['external_ip']
                      'router_public_ip': '172.24.4.23'}
        }
        self.csr = mock.Mock(spec=csr_client.CsrRestClient)
        self.csr.status = 201  # All calls to CSR REST API succeed
        self.ipsec_conn = ipsec_driver.CiscoCsrIPSecConnection(self.conn_info,
                                                               self.csr)

    def test_create_ipsec_site_connection(self):
        """Ensure all steps are done to create an IPSec site connection.

        Verify that each of the driver calls occur (in order), and
        the right information is stored for later deletion.
        """
        expected = ['create_pre_shared_key',
                    'create_ike_policy',
                    'create_ipsec_policy',
                    'create_ipsec_connection',
                    'create_static_route',
                    'create_static_route']
        expected_rollback_steps = [
            ipsec_driver.RollbackStep(action='pre_shared_key',
                                      resource_id='123',
                                      title='Pre-Shared Key'),
            ipsec_driver.RollbackStep(action='ike_policy',
                                      resource_id=222,
                                      title='IKE Policy'),
            ipsec_driver.RollbackStep(action='ipsec_policy',
                                      resource_id=333,
                                      title='IPSec Policy'),
            ipsec_driver.RollbackStep(action='ipsec_connection',
                                      resource_id='Tunnel0',
                                      title='IPSec Connection'),
            ipsec_driver.RollbackStep(action='static_route',
                                      resource_id='10.1.0.0_24_Tunnel0',
                                      title='Static Route'),
            ipsec_driver.RollbackStep(action='static_route',
                                      resource_id='10.2.0.0_24_Tunnel0',
                                      title='Static Route')]
        self.ipsec_conn.create_ipsec_site_connection(mock.Mock(),
                                                     self.conn_info)
        client_calls = [c[0] for c in self.csr.method_calls]
        self.assertEqual(expected, client_calls)
        self.assertEqual(expected_rollback_steps, self.ipsec_conn.steps)

    def test_create_ipsec_site_connection_with_rollback(self):
        """Failure test of IPSec site conn creation that fails and rolls back.

        Simulate a failure in the last create step (making routes for the
        peer networks), and ensure that the create steps are called in
        order (except for create_static_route), and that the delete
        steps are called in reverse order. At the end, there should be no
        rollback infromation for the connection.
        """
        def fake_route_check_fails(*args, **kwargs):
            if args[0] == 'Static Route':
                # So that subsequent calls to CSR rest client (for rollback)
                # will fake as passing.
                self.csr.status = httplib.NO_CONTENT
                raise ipsec_driver.CsrResourceCreateFailure(resource=args[0],
                                                            which=args[1])

        with mock.patch.object(ipsec_driver.CiscoCsrIPSecConnection,
                               '_check_create',
                               side_effect=fake_route_check_fails):

            expected = ['create_pre_shared_key',
                        'create_ike_policy',
                        'create_ipsec_policy',
                        'create_ipsec_connection',
                        'create_static_route',
                        'delete_ipsec_connection',
                        'delete_ipsec_policy',
                        'delete_ike_policy',
                        'delete_pre_shared_key']
            self.ipsec_conn.create_ipsec_site_connection(mock.Mock(),
                                                         self.conn_info)
            client_calls = [c[0] for c in self.csr.method_calls]
            self.assertEqual(expected, client_calls)
            self.assertEqual([], self.ipsec_conn.steps)

    def test_create_verification_with_error(self):
        """Negative test of create check step had failed."""
        self.csr.status = httplib.NOT_FOUND
        self.assertRaises(ipsec_driver.CsrResourceCreateFailure,
                          self.ipsec_conn._check_create, 'name', 'id')

    def test_failure_with_invalid_create_step(self):
        """Negative test of invalid create step (programming error)."""
        self.ipsec_conn.steps = []
        try:
            self.ipsec_conn.do_create_action('bogus', None, '123', 'Bad Step')
        except ipsec_driver.CsrResourceCreateFailure:
            pass
        else:
            self.fail('Expected exception with invalid create step')

    def test_failure_with_invalid_delete_step(self):
        """Negative test of invalid delete step (programming error)."""
        self.ipsec_conn.steps = [ipsec_driver.RollbackStep(action='bogus',
                                                           resource_id='123',
                                                           title='Bogus Step')]
        try:
            self.ipsec_conn.do_rollback()
        except ipsec_driver.CsrResourceCreateFailure:
            pass
        else:
            self.fail('Expected exception with invalid delete step')

    def test_delete_ipsec_connection(self):
        # TODO(pcm) implement
        pass


class TestCiscoCsrIPsecConnectionCreateTransforms(base.BaseTestCase):

    """Verifies that config info is prepared/transformed correctly."""

    def setUp(self):
        super(TestCiscoCsrIPsecConnectionCreateTransforms, self).setUp()
        self.conn_info = {
            u'id': '123',
            u'status': constants.PENDING_CREATE,
            u'admin_state_up': True,
            'psk': 'secret',
            'peer_address': '192.168.1.2',
            'peer_cidrs': ['10.1.0.0/24', '10.2.0.0/24'],
            'mtu': 1500,
            'ike_policy': {'auth_algorithm': 'sha1',
                           'encryption_algorithm': 'aes-128',
                           'pfs': 'Group5',
                           'ike_version': 'v1',
                           'lifetime_units': 'seconds',
                           'lifetime_value': 3600},
            'ipsec_policy': {'transform_protocol': 'ah',
                             'encryption_algorithm': 'aes-128',
                             'auth_algorithm': 'sha1',
                             'pfs': 'group5',
                             'lifetime_units': 'seconds',
                             'lifetime_value': 3600},
            'cisco': {'site_conn_id': 'Tunnel0',
                      'ike_policy_id': 222,
                      'ipsec_policy_id': 333,
                      # TODO(pcm) get from vpnservice['external_ip']
                      'router_public_ip': '172.24.4.23'}
        }
        self.ipsec_conn = ipsec_driver.CiscoCsrIPSecConnection(self.conn_info,
                                                               mock.Mock())

    def test_invalid_attribute(self):
        """Negative test of unknown attribute - programming error."""
        self.assertRaises(ipsec_driver.CsrDriverMismatchError,
                          self.ipsec_conn.translate_dialect,
                          'ike_policy', 'unknown_attr', self.conn_info)

    def test_driver_unknown_mapping(self):
        """Negative test of service driver providing unknown value to map."""
        self.conn_info['ike_policy']['pfs'] = "unknown_value"
        self.assertRaises(ipsec_driver.CsrUnknownMappingError,
                          self.ipsec_conn.translate_dialect,
                          'ike_policy', 'pfs', self.conn_info['ike_policy'])

    def test_psk_create_info(self):
        """Ensure that pre-shared key info is created correctly."""
        expected = {u'keyring-name': '123',
                    u'pre-shared-key-list': [
                        {u'key': 'secret',
                         u'encrypted': False,
                         u'peer-address': '192.168.1.2'}]}
        psk_id = self.conn_info['id']
        psk_info = self.ipsec_conn.create_psk_info(psk_id, self.conn_info)
        self.assertEqual(expected, psk_info)

    def test_create_ike_policy_info(self):
        """Ensure that IKE policy info is mapped/created correctly."""
        expected = {u'priority-id': 222,
                    u'encryption': u'aes',
                    u'hash': u'sha',
                    u'dhGroup': 5,
                    u'version': u'v1',
                    u'lifetime': 3600}
        policy_id = self.conn_info['cisco']['ike_policy_id']
        policy_info = self.ipsec_conn.create_ike_policy_info(policy_id,
                                                             self.conn_info)
        self.assertEqual(expected, policy_info)

    def test_create_ike_policy_info_non_defaults(self):
        """Ensure that IKE policy info with different values."""
        self.conn_info['ike_policy'] = {
            'auth_algorithm': 'sha1',
            'encryption_algorithm': 'aes-256',
            'pfs': 'Group14',
            'ike_version': 'v1',
            'lifetime_units': 'seconds',
            'lifetime_value': 60
        }
        expected = {u'priority-id': 222,
                    u'encryption': u'aes',  # TODO(pcm): fix
                    u'hash': u'sha',
                    u'dhGroup': 14,
                    u'version': u'v1',
                    u'lifetime': 60}
        policy_id = self.conn_info['cisco']['ike_policy_id']
        policy_info = self.ipsec_conn.create_ike_policy_info(policy_id,
                                                             self.conn_info)
        self.assertEqual(expected, policy_info)

    def test_ipsec_policy_info(self):
        """Ensure that IPSec policy info is mapped/created correctly."""
        expected = {u'policy-id': 333,
                    u'protection-suite': {
                        u'esp-encryption': u'esp-aes',
                        u'esp-authentication': u'esp-sha-hmac',
                        u'ah': u'ah-sha-hmac'
                    },
                    u'lifetime-sec': 3600,
                    u'pfs': u'group5',
                    u'anti-replay-window-size': u'64'}
        ipsec_policy_id = self.conn_info['cisco']['ipsec_policy_id']
        policy_info = self.ipsec_conn.create_ipsec_policy_info(ipsec_policy_id,
                                                               self.conn_info)
        self.assertEqual(expected, policy_info)

    def test_ipsec_policy_info_non_defaults(self):
        """Create/map IPSec policy info with different values."""
        self.conn_info['ipsec_policy'] = {'transform_protocol': 'esp',
                                          'encryption_algorithm': '3des',
                                          'auth_algorithm': 'sha1',
                                          'pfs': 'group14',
                                          'lifetime_units': 'seconds',
                                          'lifetime_value': 120}
        expected = {u'policy-id': 333,
                    u'protection-suite': {
                        u'esp-encryption': u'esp-3des',
                        u'esp-authentication': u'esp-sha-hmac'
                    },
                    u'lifetime-sec': 120,
                    u'pfs': u'group14',
                    u'anti-replay-window-size': u'64'}
        ipsec_policy_id = self.conn_info['cisco']['ipsec_policy_id']
        policy_info = self.ipsec_conn.create_ipsec_policy_info(ipsec_policy_id,
                                                               self.conn_info)
        self.assertEqual(expected, policy_info)

    def test_site_connection_info(self):
        """Ensure site-to-site connection info is created/mapped correctly."""
        expected = {u'vpn-interface-name': 'Tunnel0',
                    u'ipsec-policy-id': 333,
                    u'local-device': {
                        u'ip-address': u'GigabitEthernet3',
                        u'tunnel-ip-address': u'172.24.4.23'
                    },
                    u'remote-device': {
                        u'tunnel-ip-address': '192.168.1.2'
                    },
                    u'mtu': 1500}
        ipsec_policy_id = self.conn_info['cisco']['ipsec_policy_id']
        site_conn_id = self.conn_info['cisco']['site_conn_id']
        conn_info = self.ipsec_conn.create_site_connection_info(
            site_conn_id, ipsec_policy_id, self.conn_info)
        self.assertEqual(expected, conn_info)

    def test_static_route_info(self):
        """Create static route info for peer CIDRs."""
        expected = [('10.1.0.0_24_Tunnel0',
                     {u'destination-network': '10.1.0.0/24',
                      u'outgoing-interface': 'Tunnel0'}),
                    ('10.2.0.0_24_Tunnel0',
                     {u'destination-network': '10.2.0.0/24',
                      u'outgoing-interface': 'Tunnel0'})]
#         self.driver.csr.make_route_id.side_effect = ['10.1.0.0_24_Tunnel0',
#                                                      '10.2.0.0_24_Tunnel0']
        site_conn_id = self.conn_info['cisco']['site_conn_id']
        routes_info = self.ipsec_conn.create_routes_info(site_conn_id,
                                                         self.conn_info)
        self.assertEqual(2, len(routes_info))
        self.assertEqual(expected, routes_info)


class TestCiscoCsrIPsecDeviceDriverSyncStatuses(base.BaseTestCase):

    """Test status/state of services and connections, after sync."""

    def setUp(self):
        super(TestCiscoCsrIPsecDeviceDriverSyncStatuses, self).setUp()
        for klass in ['neutron.openstack.common.rpc.create_connection',
                      'neutron.context.get_admin_context_without_session',
                      'neutron.openstack.common.'
                      'loopingcall.FixedIntervalLoopingCall']:
            mock.patch(klass).start()
        self.context = context.Context('some_user', 'some_tenant')
        self.agent = mock.Mock()
        conf_patch = mock.patch('oslo.config.cfg.CONF').start()
        conf_patch.config_file = ['dummy']
        self.config_load = mock.patch(FIND_CFG_FOR_CSRS).start()
        self.config_load.return_value = {'1.1.1.1': {'rest_mgmt': '2.2.2.2',
                                                     'tunnel_ip': '1.1.1.3',
                                                     'username': 'pe',
                                                     'password': 'password',
                                                     'timeout': 120}}
        self.driver = ipsec_driver.CiscoCsrIPsecDriver(self.agent, FAKE_HOST)
        self.driver.agent_rpc = mock.Mock()
        self.conn_create = mock.patch.object(
            ipsec_driver.CiscoCsrIPSecConnection,
            'create_ipsec_site_connection').start()
        self.conn_delete = mock.patch.object(
            ipsec_driver.CiscoCsrIPSecConnection,
            'delete_ipsec_site_connection').start()
        self.csr = mock.Mock()
        self.driver.csrs['1.1.1.1'] = self.csr
        self.service123_data = {u'id': u'123',
                                u'status': constants.DOWN,
                                u'admin_state_up': False,
                                u'external_ip': u'1.1.1.1'}
        self.conn1_data = {u'id': u'1', u'status': constants.ACTIVE,
                           u'admin_state_up': True,
                           u'cisco': {u'site_conn_id': u'Tunnel0'}}

    # NOTE: For sync, there is mark (trivial), update (tested),
    # sweep (tested), and report(tested) phases.

    def test_update_ipsec_connection_create_notify(self):
        """Notified of connection create request - create."""
        # Make the (existing) service
        self.driver.create_vpn_service(self.service123_data)
        conn_data = {u'id': u'1', u'status': constants.PENDING_CREATE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}

        connection = self.driver.update_connection(self.context,
                                                   u'123', conn_data)
        self.assertFalse(connection.is_dirty)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.PENDING_CREATE, connection.last_status)
        self.assertEqual(1, self.conn_create.call_count)

    def test_update_ipsec_connection_changed_settings(self):
        """Notified of connection changing config - update."""
        # TODO(pcm) Place holder for this condition
        # Make the (existing) service and connection
        vpn_service = self.driver.create_vpn_service(self.service123_data)
        # TODO(pcm) add info that indicates that the connection has changed
        conn_data = {u'id': u'1', u'status': constants.ACTIVE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        vpn_service.create_connection(conn_data)
        self.driver.mark_existing_connections_as_dirty()

        connection = self.driver.update_connection(self.context,
                                                   '123', conn_data)
        self.assertFalse(connection.is_dirty)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.ACTIVE, connection.last_status)
        self.assertFalse(self.conn_create.called)
        # TODO(pcm) FUTURE - handling for update (delete/create?)

    def test_update_of_unknown_ipsec_connection(self):
        """Notified of update of unknown connection - create.

        Occurs if agent restarts and receives a notification of change
        to connection, but has no previous record of the connection.
        Result will be to rebuild the connection.

        This can also happen, if a connection is changed from admin
        down to admin up (so don't need a separate test for admin up.
        """
        # Will have previously created service, but don't know of connection
        self.driver.create_vpn_service(self.service123_data)
        conn_data = {u'id': u'1', u'status': constants.DOWN,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}

        connection = self.driver.update_connection(self.context,
                                                   u'123', conn_data)
        self.assertFalse(connection.is_dirty)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.DOWN, connection.last_status)
        self.assertEqual(1, self.conn_create.call_count)

    def test_update_unchanged_ipsec_connection(self):
        """Unchanged state for connection during sync - nop."""
        # Make the (existing) service and connection
        vpn_service = self.driver.create_vpn_service(self.service123_data)
        conn_data = {u'id': u'1', u'status': constants.ACTIVE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        vpn_service.create_connection(conn_data)
        self.driver.mark_existing_connections_as_dirty()
        # The notification (state) hasn't changed for the connection

        connection = self.driver.update_connection(self.context,
                                                   '123', conn_data)
        self.assertFalse(connection.is_dirty)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.ACTIVE, connection.last_status)
        self.assertFalse(self.conn_create.called)

    def test_update_connection_admin_down(self):
        """Connection updated to admin down state - force down."""
        # Make existing service, and connection that was active
        vpn_service = self.driver.create_vpn_service(self.service123_data)
        conn_data = {u'id': '1', u'status': constants.ACTIVE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        vpn_service.create_connection(conn_data)
        self.driver.mark_existing_connections_as_dirty()
        # Now simulate that the notification shows the connection admin down
        conn_data[u'admin_state_up'] = False
        conn_data[u'status'] = constants.DOWN

        connection = self.driver.update_connection(self.context,
                                                   u'123', conn_data)
        self.assertFalse(connection.is_dirty)
        self.assertTrue(connection.forced_down)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.DOWN, connection.last_status)
        self.assertFalse(self.conn_create.called)

    def test_update_missing_connection_admin_down(self):
        """Connection not present is in admin down state - nop.

        If the agent has restarted, and a sync notification occurs with
        a connection that is in admin down state, create the structures,
        but indicate that the connection is down.
        """
        # Make existing service, but no connection
        self.driver.create_vpn_service(self.service123_data)
        conn_data = {u'id': '1', u'status': constants.DOWN,
                     u'admin_state_up': False,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}

        connection = self.driver.update_connection(self.context,
                                                   u'123', conn_data)
        self.assertIsNotNone(connection)
        self.assertFalse(connection.is_dirty)
        self.assertFalse(connection.is_admin_up)
        self.assertTrue(connection.forced_down)
        self.assertFalse(self.conn_create.called)

    def test_update_connection_admin_up(self):
        """Connection updated to admin up state - record."""
        # Make existing service, and connection that was admin down
        conn_data = {u'id': '1', u'status': constants.DOWN,
                     u'admin_state_up': False,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        service_data = {u'id': u'123',
                        u'status': constants.DOWN,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn_data]}
        self.driver.update_service(self.context, service_data)
        self.driver.mark_existing_connections_as_dirty()
        # Now simulate that the notification shows the connection admin up
        conn_data[u'admin_state_up'] = True
        conn_data[u'status'] = constants.DOWN

        connection = self.driver.update_connection(self.context,
                                                   u'123', conn_data)
        self.assertFalse(connection.is_dirty)
        self.assertFalse(connection.forced_down)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.DOWN, connection.last_status)
        self.assertEqual(1, self.conn_create.call_count)

    def test_update_for_vpn_service_create(self):
        """Creation of new IPSec connection on new VPN service - create.

        Service will be created and marked as 'clean', and update
        processing for connection will occur (create).
        """
        conn_data = {u'id': u'1', u'status': constants.PENDING_CREATE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        service_data = {u'id': u'123',
                        u'status': constants.PENDING_CREATE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        self.assertFalse(vpn_service.is_dirty)
        self.assertEqual(constants.PENDING_CREATE, vpn_service.last_status)
        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        self.assertFalse(connection.is_dirty)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.PENDING_CREATE, connection.last_status)
        self.assertEqual(1, self.conn_create.call_count)

    def test_update_for_new_connection_on_existing_service(self):
        """Creating a new IPSec connection on an existing service."""
        # Create the service before testing, and mark it dirty
        prev_vpn_service = self.driver.create_vpn_service(self.service123_data)
        self.driver.mark_existing_connections_as_dirty()
        conn_data = {u'id': u'1', u'status': constants.PENDING_CREATE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        service_data = {u'id': u'123',
                        u'status': constants.ACTIVE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        # Should reuse the entry and update the status
        self.assertEqual(prev_vpn_service, vpn_service)
        self.assertFalse(vpn_service.is_dirty)
        self.assertEqual(constants.ACTIVE, vpn_service.last_status)
        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        self.assertFalse(connection.is_dirty)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.PENDING_CREATE, connection.last_status)
        self.assertEqual(1, self.conn_create.call_count)

    def test_update_for_vpn_service_with_one_unchanged_connection(self):
        """Existing VPN service and IPSec connection without any changes - nop.

        Service and connection will be marked clean. No processing for
        either, as there are no changes.
        """
        # Create a service and add in a connection that is active
        prev_vpn_service = self.driver.create_vpn_service(self.service123_data)
        conn_data = {u'id': u'1', u'status': constants.ACTIVE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        prev_vpn_service.create_connection(conn_data)
        self.driver.mark_existing_connections_as_dirty()
        # Create notification with conn unchanged and service already created
        service_data = {u'id': u'123',
                        u'status': constants.ACTIVE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        # Should reuse the entry and update the status
        self.assertEqual(prev_vpn_service, vpn_service)
        self.assertFalse(vpn_service.is_dirty)
        self.assertEqual(constants.ACTIVE, vpn_service.last_status)
        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        self.assertFalse(connection.is_dirty)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.ACTIVE, connection.last_status)
        self.assertFalse(self.conn_create.called)

    def test_update_service_admin_down(self):
        """VPN service updated to admin down state - force all down.

        If service is down, then all connections are forced down.
        """
        # Create an "existing" service, prior to notification
        prev_vpn_service = self.driver.create_vpn_service(self.service123_data)
        self.driver.mark_existing_connections_as_dirty()
        conn_data = {u'id': u'1', u'status': constants.ACTIVE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        service_data = {u'id': u'123',
                        u'status': constants.DOWN,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': False,
                        u'ipsec_conns': [conn_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        self.assertEqual(prev_vpn_service, vpn_service)
        self.assertFalse(vpn_service.is_dirty)
        self.assertFalse(vpn_service.is_admin_up)
        self.assertEqual(constants.DOWN, vpn_service.last_status)
        conn = vpn_service.get_connection(u'1')
        self.assertIsNotNone(conn)
        self.assertFalse(conn.is_dirty)
        self.assertTrue(conn.forced_down)
        self.assertTrue(conn.is_admin_up)

    def test_update_new_service_admin_down(self):
        """Unknown VPN service updated to admin down state - nop.

        Can happen if agent restarts and then gets its first notificaiton
        of a service that is in the admin down state. Structures will be
        created, but forced down.
        """
        conn_data = {u'id': u'1', u'status': constants.ACTIVE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        service_data = {u'id': u'123',
                        u'status': constants.DOWN,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': False,
                        u'ipsec_conns': [conn_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        self.assertIsNotNone(vpn_service)
        self.assertFalse(vpn_service.is_dirty)
        self.assertFalse(vpn_service.is_admin_up)
        self.assertEqual(constants.DOWN, vpn_service.last_status)
        conn = vpn_service.get_connection(u'1')
        self.assertIsNotNone(conn)
        self.assertFalse(conn.is_dirty)
        self.assertTrue(conn.forced_down)
        self.assertTrue(conn.is_admin_up)

    def test_update_service_admin_up(self):
        """VPN service updated to admin up state - restore.

        If service is up now, then connections that are admin up will come
        up and connections that are admin down, will remain down.
        """
        # Create an "existing" service, prior to notification
        prev_vpn_service = self.driver.create_vpn_service(self.service123_data)
        self.driver.mark_existing_connections_as_dirty()
        conn_data1 = {u'id': u'1', u'status': constants.DOWN,
                      u'admin_state_up': False,
                      u'cisco': {u'site_conn_id': u'Tunnel0'}}
        conn_data2 = {u'id': u'2', u'status': constants.ACTIVE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel1'}}
        service_data = {u'id': u'123',
                        u'status': constants.DOWN,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn_data1, conn_data2]}
        vpn_service = self.driver.update_service(self.context, service_data)
        self.assertEqual(prev_vpn_service, vpn_service)
        self.assertFalse(vpn_service.is_dirty)
        self.assertTrue(vpn_service.is_admin_up)
        self.assertEqual(constants.DOWN, vpn_service.last_status)
        conn1 = vpn_service.get_connection(u'1')
        self.assertIsNotNone(conn1)
        self.assertFalse(conn1.is_dirty)
        self.assertTrue(conn1.forced_down)
        self.assertFalse(conn1.is_admin_up)
        conn2 = vpn_service.get_connection(u'2')
        self.assertIsNotNone(conn2)
        self.assertFalse(conn2.is_dirty)
        self.assertFalse(conn2.forced_down)
        self.assertTrue(conn2.is_admin_up)

    def test_update_of_unknown_service_create(self):
        """Create of VPN service that is currently unknown - record.

        If agent is restarted or user changes VPN service to admin up, the
        notification may contain a VPN service with an IPSec connection
        that is not in PENDING_CREATE state.
        """
        conn_data = {u'id': u'1', u'status': constants.DOWN,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        service_data = {u'id': u'123',
                        u'status': constants.ACTIVE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        self.assertFalse(vpn_service.is_dirty)
        self.assertEqual(constants.ACTIVE, vpn_service.last_status)
        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        self.assertFalse(connection.is_dirty)
        self.assertEqual(u'Tunnel0', connection.tunnel)
        self.assertEqual(constants.DOWN, connection.last_status)
        self.assertEqual(1, self.conn_create.call_count)

    def test_update_service_create_no_csr(self):
        """Failure test of sync of service that is not on CSR - ignore.

        Ignore the VPN service and its IPSec connection(s) notifications for
        which there is no corresponding Cisco CSR.
        """
        conn_data = {u'id': u'1', u'status': constants.PENDING_CREATE,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        service_data = {u'id': u'123',
                        u'status': constants.PENDING_CREATE,
                        u'external_ip': u'2.2.2.2',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        self.assertIsNone(vpn_service)

    def _check_connection_for_service(self, count, vpn_service):
        """Helper to check the connection information for a service."""
        connection = vpn_service.get_connection(u'%d' % count)
        self.assertIsNotNone(connection, "for connection %d" % count)
        self.assertFalse(connection.is_dirty, "for connection %d" % count)
        self.assertEqual(u'Tunnel%d' % count, connection.tunnel,
                         "for connection %d" % count)
        self.assertEqual(constants.PENDING_CREATE, connection.last_status,
                         "for connection %d" % count)
        return count + 1

    def notification_for_two_services_with_two_conns(self):
        """Helper used by tests to create two services, each with two conns."""
        conn1_data = {u'id': u'1', u'status': constants.PENDING_CREATE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel1'}}
        conn2_data = {u'id': u'2', u'status': constants.PENDING_CREATE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel2'}}
        service1_data = {u'id': u'123',
                         u'status': constants.PENDING_CREATE,
                         u'external_ip': u'1.1.1.1',
                         u'admin_state_up': True,
                         u'ipsec_conns': [conn1_data, conn2_data]}
        conn3_data = {u'id': u'3', u'status': constants.PENDING_CREATE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel3'}}
        conn4_data = {u'id': u'4', u'status': constants.PENDING_CREATE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel4'}}
        service2_data = {u'id': u'456',
                         u'status': constants.PENDING_CREATE,
                         u'external_ip': u'1.1.1.1',
                         u'admin_state_up': True,
                         u'ipsec_conns': [conn3_data, conn4_data]}
        return service1_data, service2_data

    def test_create_two_connections_on_two_services(self):
        """High level test of multiple VPN services with connections."""
        # Build notification message
        (service1_data,
         service2_data) = self.notification_for_two_services_with_two_conns()
        # Simulate plugin returning notifcation, when requested
        self.driver.agent_rpc.get_vpn_services_on_host.return_value = [
            service1_data, service2_data]
        vpn_services = self.driver.update_all_services_and_connections(
            self.context)
        self.assertEqual(2, len(vpn_services))
        count = 1
        for vpn_service in vpn_services:
            self.assertFalse(vpn_service.is_dirty,
                             "for service %s" % vpn_service)
            self.assertEqual(constants.PENDING_CREATE, vpn_service.last_status,
                             "for service %s" % vpn_service)
            count = self._check_connection_for_service(count, vpn_service)
            count = self._check_connection_for_service(count, vpn_service)
        self.assertEqual(4, self.conn_create.call_count)

    def test_sweep_connection_marked_as_clean(self):
        """Sync updated connection - no action."""
        # Create a service and connection
        vpn_service = self.driver.create_vpn_service(self.service123_data)
        connection = vpn_service.create_connection(self.conn1_data)
        self.driver.mark_existing_connections_as_dirty()
        # Simulate that the update phase visted both of them
        vpn_service.is_dirty = False
        connection.is_dirty = False
        self.driver.remove_unknown_connections(self.context)
        vpn_service = self.driver.service_state.get(u'123')
        self.assertIsNotNone(vpn_service)
        self.assertFalse(vpn_service.is_dirty)
        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        self.assertFalse(connection.is_dirty)

    def test_sweep_connection_dirty(self):
        """Sync did not update connection - delete."""
        # Create a service and connection
        vpn_service = self.driver.create_vpn_service(self.service123_data)
        vpn_service.create_connection(self.conn1_data)
        self.driver.mark_existing_connections_as_dirty()
        # Simulate that the update phase only visited the service
        vpn_service.is_dirty = False
        self.driver.remove_unknown_connections(self.context)
        vpn_service = self.driver.service_state.get(u'123')
        self.assertIsNotNone(vpn_service)
        self.assertFalse(vpn_service.is_dirty)
        connection = vpn_service.get_connection(u'1')
        self.assertIsNone(connection)
        self.assertEqual(1, self.conn_delete.call_count)

    def test_sweep_service_dirty(self):
        """Sync did not update service - delete it and all conns."""
        # Create a service and connection
        vpn_service = self.driver.create_vpn_service(self.service123_data)
        vpn_service.create_connection(self.conn1_data)
        self.driver.mark_existing_connections_as_dirty()
        # Both the service and the connection are still 'dirty'
        self.driver.remove_unknown_connections(self.context)
        self.assertIsNone(self.driver.service_state.get(u'123'))
        self.assertEqual(1, self.conn_delete.call_count)

    def test_sweep_multiple_services(self):
        """One service and conn udpated, one service and conn not."""
        # Create two services, each with a connection
        vpn_service1 = self.driver.create_vpn_service(self.service123_data)
        vpn_service1.create_connection(self.conn1_data)
        service456_data = {u'id': u'456',
                           u'status': constants.ACTIVE,
                           u'admin_state_up': False,
                           u'external_ip': u'1.1.1.1'}
        conn2_data = {u'id': u'2', u'status': constants.ACTIVE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel0'}}
        prev_vpn_service2 = self.driver.create_vpn_service(service456_data)
        prev_connection2 = prev_vpn_service2.create_connection(conn2_data)
        self.driver.mark_existing_connections_as_dirty()
        # Simulate that the update phase visited the first service and conn
        prev_vpn_service2.is_dirty = False
        prev_connection2.is_dirty = False
        self.driver.remove_unknown_connections(self.context)
        self.assertIsNone(self.driver.service_state.get(u'123'))
        vpn_service2 = self.driver.service_state.get(u'456')
        self.assertEqual(prev_vpn_service2, vpn_service2)
        self.assertFalse(vpn_service2.is_dirty)
        connection2 = vpn_service2.get_connection(u'2')
        self.assertEqual(prev_connection2, connection2)
        self.assertFalse(connection2.is_dirty)
        self.assertEqual(1, self.conn_delete.call_count)

    def simulate_mark_update_sweep_for_service_with_conn(self, service_state,
                                                         connection_state):
        """Create internal structures for single service with connection.

        The service and connection will be marked as clean, and since
        none are being deleted, the service's connections_removed
        attribute will remain false.
        """
        # Simulate that we have done mark, update, and sweep.
        conn_data = {u'id': u'1', u'status': connection_state,
                     u'admin_state_up': True,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        service_data = {u'id': u'123',
                        u'status': service_state,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn_data]}
        return self.driver.update_service(self.context, service_data)

    def test_report_fragment_connection_created(self):
        """Generate report section for a created connection."""
        # Prepare service and connection in PENDING_CREATE state
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.PENDING_CREATE, constants.PENDING_CREATE)
        # Simulate that CSR has reported the connection is still up
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel0', u'UP-ACTIVE'), ]

        # Get the statuses for connections existing on CSR
        tunnels = vpn_service.get_ipsec_connections_status()
        self.assertEqual({u'Tunnel0': constants.ACTIVE}, tunnels)

        # Check that there is a status for this connection
        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        current_status = connection.find_current_status_in(tunnels)
        self.assertEqual(constants.ACTIVE, current_status)

        # Create report fragment due to change
        self.assertNotEqual(connection.last_status, current_status)
        report_frag = connection.update_status_and_build_report(current_status)
        self.assertEqual(current_status, connection.last_status)
        expected = {'1': {'status': constants.ACTIVE,
                    'updated_pending_status': True}}
        self.assertEqual(expected, report_frag)

    def test_report_fragment_connection_unchanged_status(self):
        """No report section generated for a created connection."""
        # Prepare service and connection in ACTIVE state
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.ACTIVE, constants.ACTIVE)
        # Simulate that CSR has reported the connection is up
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel0', u'UP-IDLE'), ]

        # Get the statuses for connections existing on CSR
        tunnels = vpn_service.get_ipsec_connections_status()
        self.assertEqual({u'Tunnel0': constants.ACTIVE}, tunnels)

        # Check that there is a status for this connection
        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        current_status = connection.find_current_status_in(tunnels)
        self.assertEqual(constants.ACTIVE, current_status)

        # Should be no report, as no change
        self.assertEqual(connection.last_status, current_status)
        report_frag = connection.update_status_and_build_report(current_status)
        self.assertEqual(current_status, connection.last_status)
        self.assertEqual({}, report_frag)

    def test_report_fragment_connection_changed_status(self):
        """Generate report section for connection with changed state."""
        # Prepare service in ACTIVE state and connection in DOWN state
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.ACTIVE, constants.DOWN)
        # Simulate that CSR has reported the connection is still up
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel0', u'UP-NO-IKE'), ]

        # Get the statuses for connections existing on CSR
        tunnels = vpn_service.get_ipsec_connections_status()
        self.assertEqual({u'Tunnel0': constants.ACTIVE}, tunnels)

        # Check that there is a status for this connection
        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        current_status = connection.find_current_status_in(tunnels)
        self.assertEqual(constants.ACTIVE, current_status)

        # Create report fragment due to change
        self.assertNotEqual(connection.last_status, current_status)
        report_frag = connection.update_status_and_build_report(current_status)
        self.assertEqual(current_status, connection.last_status)
        expected = {'1': {'status': constants.ACTIVE,
                    'updated_pending_status': False}}
        self.assertEqual(expected, report_frag)

    def test_report_fragment_connection_failed_create(self):
        """Failure test of report fragment for conn that failed creation.

        Normally, without any status from the CSR, the connection report would
        be skipped, but we need to report back failures.
        """
        # Prepare service and connection in PENDING_CREATE state
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.PENDING_CREATE, constants.PENDING_CREATE)
        # Simulate that CSR does NOT report the status (no tunnel)
        self.csr.read_tunnel_statuses.return_value = []

        # Get the statuses for connections existing on CSR
        tunnels = vpn_service.get_ipsec_connections_status()
        self.assertEqual({}, tunnels)

        # Check that there is a status for this connection
        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        current_status = connection.find_current_status_in(tunnels)
        self.assertEqual(constants.ERROR, current_status)

        # Create report fragment due to change
        self.assertNotEqual(connection.last_status, current_status)
        report_frag = connection.update_status_and_build_report(current_status)
        self.assertEqual(current_status, connection.last_status)
        expected = {'1': {'status': constants.ERROR,
                    'updated_pending_status': True}}
        self.assertEqual(expected, report_frag)

    def test_report_fragment_connection_admin_down(self):
        """Report for a connection that is in admin down state."""
        # Prepare service and connection with previous status ACTIVE, but
        # with connection admin down
        conn_data = {u'id': u'1', u'status': constants.ACTIVE,
                     u'admin_state_up': False,
                     u'cisco': {u'site_conn_id': u'Tunnel0'}}
        service_data = {u'id': u'123',
                        u'status': constants.ACTIVE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        # Tunnel would have been deleted, so simulate no status
        self.csr.read_tunnel_statuses.return_value = []

        connection = vpn_service.get_connection(u'1')
        self.assertIsNotNone(connection)
        self.assertTrue(connection.forced_down)
        self.assertEqual(constants.ACTIVE, connection.last_status)

        # Create report fragment due to change
        report_frag = self.driver.build_report_for_connections_on(vpn_service)
        self.assertEqual(constants.DOWN, connection.last_status)
        expected = {'1': {'status': constants.DOWN,
                    'updated_pending_status': False}}
        self.assertEqual(expected, report_frag)

    def test_report_fragment_two_connections(self):
        """Generate report fragment for two connections on a service."""
        # Prepare service with two connections, one ACTIVE, one DOWN
        conn1_data = {u'id': u'1', u'status': constants.DOWN,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel1'}}
        conn2_data = {u'id': u'2', u'status': constants.ACTIVE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel2'}}
        service_data = {u'id': u'123',
                        u'status': constants.ACTIVE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn1_data, conn2_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        # Simulate that CSR has reported the connections with diff status
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel1', u'UP-IDLE'), (u'Tunnel2', u'DOWN-NEGOTIATING')]

        # Get the report fragments for the connections
        report_frag = self.driver.build_report_for_connections_on(vpn_service)
        expected = {u'1': {u'status': constants.ACTIVE,
                           u'updated_pending_status': False},
                    u'2': {u'status': constants.DOWN,
                           u'updated_pending_status': False}}
        self.assertEqual(expected, report_frag)

    def test_report_service_create(self):
        """VPN service and IPSec connection created - report."""
        # Simulate creation of the service and connection
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.PENDING_CREATE, constants.PENDING_CREATE)
        # Simulate that the CSR has created the connection
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel0', u'UP-ACTIVE'), ]

        report = self.driver.build_report_for_service(vpn_service)
        expected_report = {
            u'id': u'123',
            u'updated_pending_status': True,
            u'status': constants.ACTIVE,
            u'ipsec_site_connections': {
                u'1': {u'status': constants.ACTIVE,
                       u'updated_pending_status': True}
            }
        }
        self.assertEqual(expected_report, report)
        # Check that service and connection statuses are updated
        self.assertEqual(constants.ACTIVE, vpn_service.last_status)
        self.assertEqual(constants.ACTIVE,
                         vpn_service.get_connection(u'1').last_status)

    def test_report_service_create_of_first_conn_fails(self):
        """VPN service and IPSec conn created, but conn failed - report.

        Since this is the sole IPSec connection on the service, and the
        create failed (connection in ERROR state), the VPN service's
        status will be set to DOWN.
        """
        # Simulate creation of the service and connection
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.PENDING_CREATE, constants.PENDING_CREATE)
        # Simulate that the CSR has no info due to failed create
        self.csr.read_tunnel_statuses.return_value = []

        report = self.driver.build_report_for_service(vpn_service)
        expected_report = {
            u'id': u'123',
            u'updated_pending_status': True,
            u'status': constants.DOWN,
            u'ipsec_site_connections': {
                u'1': {u'status': constants.ERROR,
                       u'updated_pending_status': True}
            }
        }
        self.assertEqual(expected_report, report)
        # Check that service and connection statuses are updated
        self.assertEqual(constants.DOWN, vpn_service.last_status)
        self.assertEqual(constants.ERROR,
                         vpn_service.get_connection(u'1').last_status)

    def test_report_connection_created_on_existing_service(self):
        """Creating connection on existing service - report."""
        # Simulate existing service and connection create
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.ACTIVE, constants.PENDING_CREATE)
        # Simulate that the CSR has created the connection
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel0', u'UP-IDLE'), ]

        report = self.driver.build_report_for_service(vpn_service)
        expected_report = {
            u'id': u'123',
            u'updated_pending_status': False,
            u'status': constants.ACTIVE,
            u'ipsec_site_connections': {
                u'1': {u'status': constants.ACTIVE,
                       u'updated_pending_status': True}
            }
        }
        self.assertEqual(expected_report, report)
        # Check that service and connection statuses are updated
        self.assertEqual(constants.ACTIVE, vpn_service.last_status)
        self.assertEqual(constants.ACTIVE,
                         vpn_service.get_connection(u'1').last_status)

    def test_no_report_no_changes(self):
        """VPN service with unchanged IPSec connection - no report.

        Note: No report will be generated if the last connection on the
        service is deleted. The service (and connection) objects will
        have been removed by the sweep operation and thus not reported.
        On the plugin, the service should be changed to DOWN. Likewise,
        if the service goes to admin down state.
        """
        # Simulate an existing service and connection that are ACTIVE
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.ACTIVE, constants.ACTIVE)
        # Simulate that the CSR reports the connection still active
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel0', u'UP-ACTIVE'), ]

        report = self.driver.build_report_for_service(vpn_service)
        self.assertEqual({}, report)
        # Check that service and connection statuses are still same
        self.assertEqual(constants.ACTIVE, vpn_service.last_status)
        self.assertEqual(constants.ACTIVE,
                         vpn_service.get_connection(u'1').last_status)

    def test_report_sole_connection_goes_down(self):
        """Only connection on VPN service goes down - report.

        In addition to reporting the status change and recording the new
        state for the IPSec connection, the VPN service status will be
        DOWN.
        """
        # Simulate an existing service and connection that are ACTIVE
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.ACTIVE, constants.ACTIVE)
        # Simulate that the CSR reports the connection went down
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel0', u'DOWN-NEGOTIATING'), ]

        report = self.driver.build_report_for_service(vpn_service)
        expected_report = {
            u'id': u'123',
            u'updated_pending_status': False,
            u'status': constants.DOWN,
            u'ipsec_site_connections': {
                u'1': {u'status': constants.DOWN,
                       u'updated_pending_status': False}
            }
        }
        self.assertEqual(expected_report, report)
        # Check that service and connection statuses are updated
        self.assertEqual(constants.DOWN, vpn_service.last_status)
        self.assertEqual(constants.DOWN,
                         vpn_service.get_connection(u'1').last_status)

    def test_report_sole_connection_comes_up(self):
        """Only connection on VPN service comes up - report.

        In addition to reporting the status change and recording the new
        state for the IPSec connection, the VPN service status will be
        ACTIVE.
        """
        # Simulate an existing service and connection that are DOWN
        vpn_service = self.simulate_mark_update_sweep_for_service_with_conn(
            constants.DOWN, constants.DOWN)
        # Simulate that the CSR reports the connection came up
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel0', u'UP-NO-IKE'), ]

        report = self.driver.build_report_for_service(vpn_service)
        expected_report = {
            u'id': u'123',
            u'updated_pending_status': False,
            u'status': constants.ACTIVE,
            u'ipsec_site_connections': {
                u'1': {u'status': constants.ACTIVE,
                       u'updated_pending_status': False}
            }
        }
        self.assertEqual(expected_report, report)
        # Check that service and connection statuses are updated
        self.assertEqual(constants.ACTIVE, vpn_service.last_status)
        self.assertEqual(constants.ACTIVE,
                         vpn_service.get_connection(u'1').last_status)

    def test_report_service_with_two_connections_gone_down(self):
        """One service with two connections that went down - report.

        Shows the case where all the connections are down, so that the
        service should report as DOWN, as well.
        """
        # Simulated one service with two ACTIVE connections
        conn1_data = {u'id': u'1', u'status': constants.ACTIVE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel1'}}
        conn2_data = {u'id': u'2', u'status': constants.ACTIVE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel2'}}
        service_data = {u'id': u'123',
                        u'status': constants.ACTIVE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn1_data, conn2_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        # Simulate that the CSR has reported that the connections are DOWN
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel1', u'DOWN-NEGOTIATING'), (u'Tunnel2', u'DOWN')]

        report = self.driver.build_report_for_service(vpn_service)
        expected_report = {
            u'id': u'123',
            u'updated_pending_status': False,
            u'status': constants.DOWN,
            u'ipsec_site_connections': {
                u'1': {u'status': constants.DOWN,
                       u'updated_pending_status': False},
                u'2': {u'status': constants.DOWN,
                       u'updated_pending_status': False}}
        }
        self.assertEqual(expected_report, report)
        # Check that service and connection statuses are updated
        self.assertEqual(constants.DOWN, vpn_service.last_status)
        self.assertEqual(constants.DOWN,
                         vpn_service.get_connection(u'1').last_status)
        self.assertEqual(constants.DOWN,
                         vpn_service.get_connection(u'2').last_status)

    def test_report_service_with_connection_removed(self):
        """One service with two connections where one is removed - report.

        With a connection removed and the other connection unchanged,
        normally there would be nothing to report for the connections, but
        we need to report any possible change to the service state. In this
        case, the service was ACTIVE, but since the only ACTIVE connection
        is deleted and the remaining connection is DOWN, the service will
        indicate as DOWN.
        """
        # Simulated one service with one connection up, one down
        conn1_data = {u'id': u'1', u'status': constants.ACTIVE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel1'}}
        conn2_data = {u'id': u'2', u'status': constants.DOWN,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel2'}}
        service_data = {u'id': u'123',
                        u'status': constants.ACTIVE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn1_data, conn2_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        self.assertEqual(constants.ACTIVE, vpn_service.last_status)
        self.assertEqual(constants.ACTIVE,
                         vpn_service.get_connection(u'1').last_status)
        self.assertEqual(constants.DOWN,
                         vpn_service.get_connection(u'2').last_status)

        # Simulate that one is deleted
        self.driver.mark_existing_connections_as_dirty()
        service_data = {u'id': u'123',
                        u'status': constants.ACTIVE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': True,
                        u'ipsec_conns': [conn2_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        self.driver.remove_unknown_connections(self.context)
        self.assertTrue(vpn_service.connections_removed)
        self.assertEqual(constants.ACTIVE, vpn_service.last_status)
        self.assertIsNone(vpn_service.get_connection(u'1'))
        self.assertEqual(constants.DOWN,
                         vpn_service.get_connection(u'2').last_status)

        # Simulate that only one connection reports and status is unchanged,
        # so there will be NO connection info to report.
        self.csr.read_tunnel_statuses.return_value = [(u'Tunnel2', u'DOWN')]
        report = self.driver.build_report_for_service(vpn_service)
        expected_report = {
            u'id': u'123',
            u'updated_pending_status': False,
            u'status': constants.DOWN,
            u'ipsec_site_connections': {}
        }
        self.assertEqual(expected_report, report)
        # Check that service and connection statuses are updated
        self.assertEqual(constants.DOWN, vpn_service.last_status)
        self.assertEqual(constants.DOWN,
                         vpn_service.get_connection(u'2').last_status)

    def test_report_service_admin_down_with_two_connections(self):
        """One service admin down, with two connections - report.

        When the service is admin down, all the connections will report
        as DOWN.
        """
        # Simulated one service (admin down) with two ACTIVE connections
        conn1_data = {u'id': u'1', u'status': constants.ACTIVE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel1'}}
        conn2_data = {u'id': u'2', u'status': constants.ACTIVE,
                      u'admin_state_up': True,
                      u'cisco': {u'site_conn_id': u'Tunnel2'}}
        service_data = {u'id': u'123',
                        u'status': constants.ACTIVE,
                        u'external_ip': u'1.1.1.1',
                        u'admin_state_up': False,
                        u'ipsec_conns': [conn1_data, conn2_data]}
        vpn_service = self.driver.update_service(self.context, service_data)
        # Since service admin down, connections will have been deleted
        self.csr.read_tunnel_statuses.return_value = []

        report = self.driver.build_report_for_service(vpn_service)
        expected_report = {
            u'id': u'123',
            u'updated_pending_status': False,
            u'status': constants.DOWN,
            u'ipsec_site_connections': {
                u'1': {u'status': constants.DOWN,
                       u'updated_pending_status': False},
                u'2': {u'status': constants.DOWN,
                       u'updated_pending_status': False}}
        }
        self.assertEqual(expected_report, report)
        # Check that service and connection statuses are updated
        self.assertEqual(constants.DOWN, vpn_service.last_status)
        self.assertEqual(constants.DOWN,
                         vpn_service.get_connection(u'1').last_status)
        self.assertEqual(constants.DOWN,
                         vpn_service.get_connection(u'2').last_status)

    def test_report_multiple_services(self):
        """Status changes for several services - report."""
        # Simulate creation of the service and connection
        (service1_data,
         service2_data) = self.notification_for_two_services_with_two_conns()
        vpn_service1 = self.driver.update_service(self.context, service1_data)
        vpn_service2 = self.driver.update_service(self.context, service2_data)
        # Simulate that the CSR has created the connections
        self.csr.read_tunnel_statuses.return_value = [
            (u'Tunnel1', u'UP-ACTIVE'), (u'Tunnel2', u'DOWN'),
            (u'Tunnel3', u'DOWN-NEGOTIATING'), (u'Tunnel4', u'UP-IDLE')]

        report = self.driver.report_status(self.context)
        expected_report = [{u'id': u'123',
                            u'updated_pending_status': True,
                            u'status': constants.ACTIVE,
                            u'ipsec_site_connections': {
                                u'1': {u'status': constants.ACTIVE,
                                       u'updated_pending_status': True},
                            u'2': {u'status': constants.DOWN,
                                   u'updated_pending_status': True}}
                            },
                           {u'id': u'456',
                            u'updated_pending_status': True,
                            u'status': constants.ACTIVE,
                            u'ipsec_site_connections': {
                                u'3': {u'status': constants.DOWN,
                                       u'updated_pending_status': True},
                                u'4': {u'status': constants.ACTIVE,
                                       u'updated_pending_status': True}}
                            }]
        self.assertEqual(expected_report, report)
        # Check that service and connection statuses are updated
        self.assertEqual(constants.ACTIVE, vpn_service1.last_status)
        self.assertEqual(constants.ACTIVE,
                         vpn_service1.get_connection(u'1').last_status)
        self.assertEqual(constants.DOWN,
                         vpn_service1.get_connection(u'2').last_status)
        self.assertEqual(constants.ACTIVE, vpn_service2.last_status)
        self.assertEqual(constants.DOWN,
                         vpn_service2.get_connection(u'3').last_status)
        self.assertEqual(constants.ACTIVE,
                         vpn_service2.get_connection(u'4').last_status)

    # TODO(pcm) FUTURE - UTs for update action, when supported.

    def test_vpnservice_updated(self):
        with mock.patch.object(self.driver, 'sync') as sync:
            context = mock.Mock()
            self.driver.vpnservice_updated(context)
            sync.assert_called_once_with(context, [])


class TestCiscoCsrIPsecDeviceDriverConfigLoading(base.BaseTestCase):

    def create_tempfile(self, contents):
        (fd, path) = tempfile.mkstemp(prefix='test', suffix='.conf')
        try:
            os.write(fd, contents.encode('utf-8'))
        finally:
            os.close(fd)
        return path

    def test_loading_csr_configuration(self):
        """Ensure that Cisco CSR configs can be loaded from config files."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:3.2.1.1]\n'
                                        'rest_mgmt = 10.20.30.1\n'
                                        'tunnel_ip = 3.2.1.3\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = 5.0\n')
        expected = {'3.2.1.1': {'rest_mgmt': '10.20.30.1',
                                'tunnel_ip': '3.2.1.3',
                                'username': 'me',
                                'password': 'secret',
                                'timeout': 5.0}}
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual(expected, csrs_found)

    def test_loading_config_without_timeout(self):
        """Cisco CSR config without timeout will use default timeout."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:3.2.1.1]\n'
                                        'rest_mgmt = 10.20.30.1\n'
                                        'tunnel_ip = 3.2.1.3\n'
                                        'username = me\n'
                                        'password = secret\n')
        expected = {'3.2.1.1': {'rest_mgmt': '10.20.30.1',
                                'tunnel_ip': '3.2.1.3',
                                'username': 'me',
                                'password': 'secret',
                                'timeout': csr_client.TIMEOUT}}
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual(expected, csrs_found)

    def test_skip_loading_duplicate_csr_configuration(self):
        """Failure test that duplicate configurations are ignored."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:3.2.1.1]\n'
                                        'rest_mgmt = 10.20.30.1\n'
                                        'tunnel_ip = 3.2.1.3\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = 5.0\n'
                                        '[CISCO_CSR_REST:3.2.1.1]\n'
                                        'rest_mgmt = 5.5.5.3\n'
                                        'tunnel_ip = 3.2.1.6\n'
                                        'username = me\n'
                                        'password = secret\n')
        expected = {'3.2.1.1': {'rest_mgmt': '10.20.30.1',
                                'tunnel_ip': '3.2.1.3',
                                'username': 'me',
                                'password': 'secret',
                                'timeout': 5.0}}
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual(expected, csrs_found)

    def test_fail_loading_config_with_invalid_timeout(self):
        """Failure test of invalid timeout in config info."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:3.2.1.1]\n'
                                        'rest_mgmt = 10.20.30.1\n'
                                        'tunnel_ip = 3.2.1.3\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = yes\n')
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual({}, csrs_found)

    def test_fail_loading_config_missing_required_info(self):
        """Failure test of config missing required info."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:1.1.1.0]\n'
                                        'tunnel_ip = 1.1.1.3\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = 5.0\n'
                                        '[CISCO_CSR_REST:2.2.2.0]\n'
                                        'rest_mgmt = 10.20.30.1\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = 5.0\n'
                                        '[CISCO_CSR_REST:3.3.3.0]\n'
                                        'rest_mgmt = 10.20.30.1\n'
                                        'tunnel_ip = 3.3.3.3\n'
                                        'password = secret\n'
                                        'timeout = 5.0\n'
                                        '[CISCO_CSR_REST:4.4.4.0]\n'
                                        'rest_mgmt = 10.20.30.1\n'
                                        'tunnel_ip = 4.4.4.4\n'
                                        'username = me\n'
                                        'timeout = 5.0\n')
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual({}, csrs_found)

    def test_fail_loading_config_with_invalid_router_id(self):
        """Failure test of config with invalid rotuer ID."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:4.3.2.1.9]\n'
                                        'rest_mgmt = 10.20.30.1\n'
                                        'tunnel_ip = 4.3.2.3\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = 5.0\n')
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual({}, csrs_found)

    def test_fail_loading_config_with_invalid_mgmt_ip(self):
        """Failure test of configuration with invalid management IP address."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:3.2.1.1]\n'
                                        'rest_mgmt = 1.1.1.1.1\n'
                                        'tunnel_ip = 3.2.1.3\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = 5.0\n')
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual({}, csrs_found)

    def test_fail_loading_config_with_invalid_tunnel_ip(self):
        """Failure test of configuration with invalid tunnel IP address."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:3.2.1.1]\n'
                                        'rest_mgmt = 1.1.1.1\n'
                                        'tunnel_ip = 3.2.1.4.5\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = 5.0\n')
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual({}, csrs_found)

    def test_failure_no_configurations_entries(self):
        """Failure test config file without any CSR definitions."""
        cfg_file = self.create_tempfile('NO CISCO SECTION AT ALL\n')
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual({}, csrs_found)

    def test_failure_no_csr_configurations_entries(self):
        """Failure test config file without any CSR definitions."""
        cfg_file = self.create_tempfile('[SOME_CONFIG:123]\n'
                                        'username = me\n')
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual({}, csrs_found)

    def test_missing_config_value(self):
        """Failure test of config file missing a value for attribute."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:3.2.1.1]\n'
                                        'rest_mgmt = \n'
                                        'tunnel_ip = 3.2.1.3\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = 5.0\n')
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual({}, csrs_found)

    def test_ignores_invalid_attribute_in_config(self):
        """Test ignoring of config file with invalid attribute."""
        cfg_file = self.create_tempfile('[CISCO_CSR_REST:3.2.1.1]\n'
                                        'rest_mgmt = 1.1.1.1\n'
                                        'bogus = abcdef\n'
                                        'tunnel_ip = 3.2.1.3\n'
                                        'username = me\n'
                                        'password = secret\n'
                                        'timeout = 15.5\n')
        expected = {'3.2.1.1': {'rest_mgmt': '1.1.1.1',
                                'tunnel_ip': '3.2.1.3',
                                'username': 'me',
                                'password': 'secret',
                                'timeout': 15.5}}
        csrs_found = ipsec_driver.find_available_csrs_from_config([cfg_file])
        self.assertEqual(expected, csrs_found)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import time

from oslo.config import cfg
from webob import exc

from neutron.api.v2 import attributes
from neutron.common import constants
from neutron.common import topics
from neutron import context
from neutron.db import agents_db
from neutron.db import db_base_plugin_v2
from neutron.extensions import agent
from neutron.openstack.common import log as logging
from neutron.openstack.common import timeutils
from neutron.openstack.common import uuidutils
from neutron.tests.unit import test_api_v2
from neutron.tests.unit import test_db_plugin


LOG = logging.getLogger(__name__)

_uuid = uuidutils.generate_uuid
_get_path = test_api_v2._get_path
L3_HOSTA = 'hosta'
DHCP_HOSTA = 'hosta'
L3_HOSTB = 'hostb'
DHCP_HOSTC = 'hostc'
DHCP_HOST1 = 'host1'
LBAAS_HOSTA = 'hosta'
LBAAS_HOSTB = 'hostb'


class AgentTestExtensionManager(object):

    def get_resources(self):
        # Add the resources to the global attribute map
        # This is done here as the setup process won't
        # initialize the main API router which extends
        # the global attribute map
        attributes.RESOURCE_ATTRIBUTE_MAP.update(
            agent.RESOURCE_ATTRIBUTE_MAP)
        return agent.Agent.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


# This plugin class is just for testing
class TestAgentPlugin(db_base_plugin_v2.NeutronDbPluginV2,
                      agents_db.AgentDbMixin):
    supported_extension_aliases = ["agent"]


class AgentDBTestMixIn(object):

    def _list_agents(self, expected_res_status=None,
                     neutron_context=None,
                     query_string=None):
        agent_res = self._list('agents',
                               neutron_context=neutron_context,
                               query_params=query_string)
        if expected_res_status:
            self.assertEqual(agent_res.status_int, expected_res_status)
        return agent_res

    def _register_agent_states(self, lbaas_agents=False):
        """Register two L3 agents and two DHCP agents."""
        l3_hosta = {
            'binary': 'neutron-l3-agent',
            'host': L3_HOSTA,
            'topic': topics.L3_AGENT,
            'configurations': {'use_namespaces': True,
                               'router_id': None,
                               'handle_internal_only_routers':
                               True,
                               'gateway_external_network_id':
                               None,
                               'interface_driver': 'interface_driver',
                               },
            'agent_type': constants.AGENT_TYPE_L3}
        l3_hostb = copy.deepcopy(l3_hosta)
        l3_hostb['host'] = L3_HOSTB
        dhcp_hosta = {
            'binary': 'neutron-dhcp-agent',
            'host': DHCP_HOSTA,
            'topic': 'DHCP_AGENT',
            'configurations': {'dhcp_driver': 'dhcp_driver',
                               'use_namespaces': True,
                               },
            'agent_type': constants.AGENT_TYPE_DHCP}
        dhcp_hostc = copy.deepcopy(dhcp_hosta)
        dhcp_hostc['host'] = DHCP_HOSTC
        lbaas_hosta = {
            'binary': 'neutron-loadbalancer-agent',
            'host': LBAAS_HOSTA,
            'topic': 'LOADBALANCER_AGENT',
            'configurations': {'device_drivers': ['haproxy_ns']},
            'agent_type': constants.AGENT_TYPE_LOADBALANCER}
        lbaas_hostb = copy.deepcopy(lbaas_hosta)
        lbaas_hostb['host'] = LBAAS_HOSTB
        callback = agents_db.AgentExtRpcCallback()
        callback.report_state(self.adminContext,
                              agent_state={'agent_state': l3_hosta},
                              time=timeutils.strtime())
        callback.report_state(self.adminContext,
                              agent_state={'agent_state': l3_hostb},
                              time=timeutils.strtime())
        callback.report_state(self.adminContext,
                              agent_state={'agent_state': dhcp_hosta},
                              time=timeutils.strtime())
        callback.report_state(self.adminContext,
                              agent_state={'agent_state': dhcp_hostc},
                              time=timeutils.strtime())

        res = [l3_hosta, l3_hostb, dhcp_hosta, dhcp_hostc]
        if lbaas_agents:
            callback.report_state(self.adminContext,
                                  agent_state={'agent_state': lbaas_hosta},
                                  time=timeutils.strtime())
            callback.report_state(self.adminContext,
                                  agent_state={'agent_state': lbaas_hostb},
                                  time=timeutils.strtime())
            res += [lbaas_hosta, lbaas_hostb]

        return res

    def _register_one_dhcp_agent(self):
        """Register one DHCP agent."""
        dhcp_host = {
            'binary': 'neutron-dhcp-agent',
            'host': DHCP_HOST1,
            'topic': 'DHCP_AGENT',
            'configurations': {'dhcp_driver': 'dhcp_driver',
                               'use_namespaces': True,
                               },
            'agent_type': constants.AGENT_TYPE_DHCP}
        callback = agents_db.AgentExtRpcCallback()
        callback.report_state(self.adminContext,
                              agent_state={'agent_state': dhcp_host},
                              time=timeutils.strtime())
        return [dhcp_host]


class AgentDBTestCase(AgentDBTestMixIn,
                      test_db_plugin.NeutronDbPluginV2TestCase):
    fmt = 'json'

    def setUp(self):
        self.adminContext = context.get_admin_context()
        plugin = 'neutron.tests.unit.test_agent_ext_plugin.TestAgentPlugin'
        # for these tests we need to enable overlapping ips
        cfg.CONF.set_default('allow_overlapping_ips', True)
        # Save the original RESOURCE_ATTRIBUTE_MAP
        self.saved_attr_map = {}
        for resource, attrs in attributes.RESOURCE_ATTRIBUTE_MAP.iteritems():
            self.saved_attr_map[resource] = attrs.copy()
        ext_mgr = AgentTestExtensionManager()
        self.addCleanup(self.restore_resource_attribute_map)
        super(AgentDBTestCase, self).setUp(plugin=plugin, ext_mgr=ext_mgr)

    def restore_resource_attribute_map(self):
        # Restore the originak RESOURCE_ATTRIBUTE_MAP
        attributes.RESOURCE_ATTRIBUTE_MAP = self.saved_attr_map

    def test_create_agent(self):
        data = {'agent': {}}
        _req = self.new_create_request('agents', data, self.fmt)
        _req.environ['neutron.context'] = context.Context(
            '', 'tenant_id')
        res = _req.get_response(self.ext_api)
        self.assertEqual(res.status_int, exc.HTTPBadRequest.code)

    def test_list_agent(self):
        agents = self._register_agent_states()
        res = self._list('agents')
        for agent in res['agents']:
            if (agent['host'] == DHCP_HOSTA and
                agent['agent_type'] == constants.AGENT_TYPE_DHCP):
                self.assertEqual(
                    'dhcp_driver',
                    agent['configurations']['dhcp_driver'])
                break
        self.assertEqual(len(agents), len(res['agents']))

    def test_show_agent(self):
        self._register_agent_states()
        agents = self._list_agents(
            query_string='binary=neutron-l3-agent')
        self.assertEqual(2, len(agents['agents']))
        agent = self._show('agents', agents['agents'][0]['id'])
        self.assertEqual('neutron-l3-agent', agent['agent']['binary'])

    def test_update_agent(self):
        self._register_agent_states()
        agents = self._list_agents(
            query_string='binary=neutron-l3-agent&host=' + L3_HOSTB)
        self.assertEqual(1, len(agents['agents']))
        com_id = agents['agents'][0]['id']
        agent = self._show('agents', com_id)
        new_agent = {}
        new_agent['agent'] = {}
        new_agent['agent']['admin_state_up'] = False
        new_agent['agent']['description'] = 'description'
        self._update('agents', com_id, new_agent)
        agent = self._show('agents', com_id)
        self.assertFalse(agent['agent']['admin_state_up'])
        self.assertEqual('description', agent['agent']['description'])

    def test_dead_agent(self):
        cfg.CONF.set_override('agent_down_time', 1)
        self._register_agent_states()
        time.sleep(1.5)
        agents = self._list_agents(
            query_string='binary=neutron-l3-agent&host=' + L3_HOSTB)
        self.assertFalse(agents['agents'][0]['alive'])


class AgentDBTestCaseXML(AgentDBTestCase):
    fmt = 'xml'



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2012, Nachi Ueno, NTT MCL, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

import mock

from neutron.api.v2 import attributes
from neutron.extensions import securitygroup as ext_sg
from neutron import manager
from neutron.tests.unit.ryu import fake_ryu
from neutron.tests.unit import test_extension_security_group as test_sg
from neutron.tests.unit import test_security_groups_rpc as test_sg_rpc

PLUGIN_NAME = ('neutron.plugins.ryu.'
               'ryu_neutron_plugin.RyuNeutronPluginV2')
NOTIFIER = ('neutron.plugins.ryu.'
            'ryu_neutron_plugin.AgentNotifierApi')


class RyuSecurityGroupsTestCase(test_sg.SecurityGroupDBTestCase):
    _plugin_name = PLUGIN_NAME

    def setUp(self, plugin=None):
        test_sg_rpc.set_firewall_driver(test_sg_rpc.FIREWALL_HYBRID_DRIVER)
        self.fake_ryu = fake_ryu.patch_fake_ryu_client().start()
        notifier_p = mock.patch(NOTIFIER)
        notifier_cls = notifier_p.start()
        self.notifier = mock.Mock()
        notifier_cls.return_value = self.notifier
        self._attribute_map_bk_ = {}
        for item in attributes.RESOURCE_ATTRIBUTE_MAP:
            self._attribute_map_bk_[item] = (attributes.
                                             RESOURCE_ATTRIBUTE_MAP[item].
                                             copy())
        super(RyuSecurityGroupsTestCase, self).setUp(PLUGIN_NAME)

    def tearDown(self):
        super(RyuSecurityGroupsTestCase, self).tearDown()
        attributes.RESOURCE_ATTRIBUTE_MAP = self._attribute_map_bk_


class TestRyuSecurityGroups(RyuSecurityGroupsTestCase,
                            test_sg.TestSecurityGroups,
                            test_sg_rpc.SGNotificationTestMixin):
    def test_security_group_get_port_from_device(self):
        with contextlib.nested(self.network(),
                               self.security_group()) as (n, sg):
            with self.subnet(n):
                security_group_id = sg['security_group']['id']
                res = self._create_port(self.fmt, n['network']['id'])
                port = self.deserialize(self.fmt, res)
                fixed_ips = port['port']['fixed_ips']
                data = {'port': {'fixed_ips': fixed_ips,
                                 'name': port['port']['name'],
                                 ext_sg.SECURITYGROUPS:
                                 [security_group_id]}}

                req = self.new_update_request('ports', data,
                                              port['port']['id'])
                res = self.deserialize(self.fmt,
                                       req.get_response(self.api))
                port_id = res['port']['id']
                plugin = manager.NeutronManager.get_plugin()
                port_dict = plugin.callbacks.get_port_from_device(port_id)
                self.assertEqual(port_id, port_dict['id'])
                self.assertEqual([security_group_id],
                                 port_dict[ext_sg.SECURITYGROUPS])
                self.assertEqual([], port_dict['security_group_rules'])
                self.assertEqual([fixed_ips[0]['ip_address']],
                                 port_dict['fixed_ips'])
                self._delete('ports', port_id)

    def test_security_group_get_port_from_device_with_no_port(self):
        plugin = manager.NeutronManager.get_plugin()
        port_dict = plugin.callbacks.get_port_from_device('bad_device_id')
        self.assertIsNone(port_dict)


class TestRyuSecurityGroupsXML(TestRyuSecurityGroups):
    fmt = 'xml'



# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Author: Sylvain Afchain <sylvain.afchain@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib
import logging

import webob.exc

from neutron.api.extensions import ExtensionMiddleware
from neutron.api.extensions import PluginAwareExtensionManager
from neutron.common import config
from neutron import context
import neutron.extensions
from neutron.extensions import metering
from neutron.plugins.common import constants
from neutron.services.metering import metering_plugin
from neutron.tests.unit import test_db_plugin

LOG = logging.getLogger(__name__)

DB_METERING_PLUGIN_KLASS = (
    "neutron.services.metering."
    "metering_plugin.MeteringPlugin"
)

extensions_path = ':'.join(neutron.extensions.__path__)


class MeteringPluginDbTestCaseMixin(object):
    def _create_metering_label(self, fmt, name, description, **kwargs):
        data = {'metering_label': {'name': name,
                                   'tenant_id': kwargs.get('tenant_id',
                                                           'test_tenant'),
                                   'description': description}}
        req = self.new_create_request('metering-labels', data,
                                      fmt)

        if kwargs.get('set_context') and 'tenant_id' in kwargs:
            # create a specific auth context for this request
            req.environ['neutron.context'] = (
                context.Context('', kwargs['tenant_id'],
                                is_admin=kwargs.get('is_admin', True)))

        return req.get_response(self.ext_api)

    def _make_metering_label(self, fmt, name, description, **kwargs):
        res = self._create_metering_label(fmt, name, description, **kwargs)
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(fmt, res)

    def _create_metering_label_rule(self, fmt, metering_label_id, direction,
                                    remote_ip_prefix, excluded, **kwargs):
        data = {'metering_label_rule':
                {'metering_label_id': metering_label_id,
                 'tenant_id': kwargs.get('tenant_id', 'test_tenant'),
                 'direction': direction,
                 'excluded': excluded,
                 'remote_ip_prefix': remote_ip_prefix}}
        req = self.new_create_request('metering-label-rules',
                                      data, fmt)

        if kwargs.get('set_context') and 'tenant_id' in kwargs:
            # create a specific auth context for this request
            req.environ['neutron.context'] = (
                context.Context('', kwargs['tenant_id']))

        return req.get_response(self.ext_api)

    def _make_metering_label_rule(self, fmt, metering_label_id, direction,
                                  remote_ip_prefix, excluded, **kwargs):
        res = self._create_metering_label_rule(fmt, metering_label_id,
                                               direction, remote_ip_prefix,
                                               excluded, **kwargs)
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(fmt, res)

    @contextlib.contextmanager
    def metering_label(self, name='label', description='desc',
                       fmt=None, no_delete=False, **kwargs):
        if not fmt:
            fmt = self.fmt
        metering_label = self._make_metering_label(fmt, name,
                                                   description, **kwargs)
        yield metering_label
        if not no_delete:
            self._delete('metering-labels',
                         metering_label['metering_label']['id'])

    @contextlib.contextmanager
    def metering_label_rule(self, metering_label_id=None, direction='ingress',
                            remote_ip_prefix='10.0.0.0/24',
                            excluded='false', fmt=None, no_delete=False):
        if not fmt:
            fmt = self.fmt
        metering_label_rule = self._make_metering_label_rule(fmt,
                                                             metering_label_id,
                                                             direction,
                                                             remote_ip_prefix,
                                                             excluded)
        yield metering_label_rule
        if not no_delete:
            self._delete('metering-label-rules',
                         metering_label_rule['metering_label_rule']['id'])


class MeteringPluginDbTestCase(test_db_plugin.NeutronDbPluginV2TestCase,
                               MeteringPluginDbTestCaseMixin):
    fmt = 'json'

    resource_prefix_map = dict(
        (k.replace('_', '-'), constants.COMMON_PREFIXES[constants.METERING])
        for k in metering.RESOURCE_ATTRIBUTE_MAP.keys()
    )

    def setUp(self, plugin=None):
        service_plugins = {'metering_plugin_name': DB_METERING_PLUGIN_KLASS}

        super(MeteringPluginDbTestCase, self).setUp(
            plugin=plugin,
            service_plugins=service_plugins
        )

        self.plugin = metering_plugin.MeteringPlugin()
        ext_mgr = PluginAwareExtensionManager(
            extensions_path,
            {constants.METERING: self.plugin}
        )
        app = config.load_paste_app('extensions_test_app')
        self.ext_api = ExtensionMiddleware(app, ext_mgr=ext_mgr)

    def test_create_metering_label(self):
        name = 'my label'
        description = 'my metering label'
        keys = [('name', name,), ('description', description)]
        with self.metering_label(name, description) as metering_label:
            for k, v, in keys:
                self.assertEqual(metering_label['metering_label'][k], v)

    def test_delete_metering_label(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description,
                                 no_delete=True) as metering_label:
            metering_label_id = metering_label['metering_label']['id']
            self._delete('metering-labels', metering_label_id, 204)

    def test_list_metering_label(self):
        name = 'my label'
        description = 'my metering label'

        with contextlib.nested(
                self.metering_label(name, description),
                self.metering_label(name, description)) as metering_label:

            self._test_list_resources('metering-label', metering_label)

    def test_create_metering_label_rule(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']

            direction = 'egress'
            remote_ip_prefix = '192.168.0.0/24'
            excluded = True

            keys = [('metering_label_id', metering_label_id),
                    ('direction', direction),
                    ('excluded', excluded),
                    ('remote_ip_prefix', remote_ip_prefix)]
            with self.metering_label_rule(metering_label_id,
                                          direction,
                                          remote_ip_prefix,
                                          excluded) as label_rule:
                for k, v, in keys:
                    self.assertEqual(label_rule['metering_label_rule'][k], v)

    def test_delete_metering_label_rule(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']

            direction = 'egress'
            remote_ip_prefix = '192.168.0.0/24'
            excluded = True

            with self.metering_label_rule(metering_label_id,
                                          direction,
                                          remote_ip_prefix,
                                          excluded,
                                          no_delete=True) as label_rule:
                rule_id = label_rule['metering_label_rule']['id']
                self._delete('metering-label-rules', rule_id, 204)

    def test_list_metering_label_rule(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']

            direction = 'egress'
            remote_ip_prefix = '192.168.0.0/24'
            excluded = True

            with contextlib.nested(
                self.metering_label_rule(metering_label_id,
                                         direction,
                                         remote_ip_prefix,
                                         excluded),
                self.metering_label_rule(metering_label_id,
                                         'ingress',
                                         remote_ip_prefix,
                                         excluded)) as metering_label_rule:

                self._test_list_resources('metering-label-rule',
                                          metering_label_rule)

    def test_create_metering_label_rules(self):
        name = 'my label'
        description = 'my metering label'

        with self.metering_label(name, description) as metering_label:
            metering_label_id = metering_label['metering_label']['id']

            direction = 'egress'
            remote_ip_prefix = '192.168.0.0/24'
            excluded = True

            with contextlib.nested(
                self.metering_label_rule(metering_label_id,
                                         direction,
                                         remote_ip_prefix,
                                         excluded),
                self.metering_label_rule(metering_label_id,
                                         direction,
                                         '0.0.0.0/0',
                                         False)) as metering_label_rule:

                self._test_list_resources('metering-label-rule',
                                          metering_label_rule)

    def test_create_metering_label_rule_two_labels(self):
        name1 = 'my label 1'
        name2 = 'my label 2'
        description = 'my metering label'

        with self.metering_label(name1, description) as metering_label1:
            metering_label_id1 = metering_label1['metering_label']['id']

            with self.metering_label(name2, description) as metering_label2:
                metering_label_id2 = metering_label2['metering_label']['id']

                direction = 'egress'
                remote_ip_prefix = '192.168.0.0/24'
                excluded = True

                with contextlib.nested(
                    self.metering_label_rule(metering_label_id1,
                                             direction,
                                             remote_ip_prefix,
                                             excluded),
                    self.metering_label_rule(metering_label_id2,
                                             direction,
                                             remote_ip_prefix,
                                             excluded)) as metering_label_rule:

                    self._test_list_resources('metering-label-rule',
                                              metering_label_rule)


class TestMeteringDbXML(MeteringPluginDbTestCase):
    fmt = 'xml'



# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Author: Sylvain Afchain <sylvain.afchain@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo.config import cfg

from neutron.agent.common import config
from neutron.agent.linux import interface
from neutron.agent.linux import iptables_manager
from neutron.common import constants as constants
from neutron.common import ipv6_utils
from neutron.common import log
from neutron.openstack.common import importutils
from neutron.openstack.common import log as logging
from neutron.services.metering.drivers import abstract_driver


LOG = logging.getLogger(__name__)
NS_PREFIX = 'qrouter-'
WRAP_NAME = 'neutron-meter'
EXTERNAL_DEV_PREFIX = 'qg-'
TOP_CHAIN = WRAP_NAME + "-FORWARD"
RULE = '-r-'
LABEL = '-l-'

config.register_interface_driver_opts_helper(cfg.CONF)
config.register_use_namespaces_opts_helper(cfg.CONF)
config.register_root_helper(cfg.CONF)
cfg.CONF.register_opts(interface.OPTS)


class IptablesManagerTransaction(object):
    __transactions = {}

    def __init__(self, im):
        self.im = im

        transaction = self.__transactions.get(im, 0)
        transaction += 1
        self.__transactions[im] = transaction

    def __enter__(self):
        return self.im

    def __exit__(self, type, value, traceback):
        transaction = self.__transactions.get(self.im)
        if transaction == 1:
            self.im.apply()
            del self.__transactions[self.im]
        else:
            transaction -= 1
            self.__transactions[self.im] = transaction


class RouterWithMetering(object):

    def __init__(self, conf, router):
        self.conf = conf
        self.id = router['id']
        self.router = router
        self.root_helper = config.get_root_helper(self.conf)
        self.ns_name = NS_PREFIX + self.id if conf.use_namespaces else None
        self.iptables_manager = iptables_manager.IptablesManager(
            root_helper=self.root_helper,
            namespace=self.ns_name,
            binary_name=WRAP_NAME,
            use_ipv6=ipv6_utils.is_enabled())
        self.metering_labels = {}


class IptablesMeteringDriver(abstract_driver.MeteringAbstractDriver):

    def __init__(self, plugin, conf):
        self.plugin = plugin
        self.conf = conf or cfg.CONF
        self.routers = {}

        if not self.conf.interface_driver:
            raise SystemExit(_('An interface driver must be specified'))
        LOG.info(_("Loading interface driver %s"), self.conf.interface_driver)
        self.driver = importutils.import_object(self.conf.interface_driver,
                                                self.conf)

    def _update_router(self, router):
        r = self.routers.get(router['id'],
                             RouterWithMetering(self.conf, router))
        r.router = router
        self.routers[r.id] = r

        return r

    @log.log
    def update_routers(self, context, routers):
        # disassociate removed routers
        router_ids = [router['id'] for router in routers]
        for router_id in self.routers:
            if router_id not in router_ids:
                self._process_disassociate_metering_label(router)

        for router in routers:
            old_gw_port_id = None
            old_rm = self.routers.get(router['id'])
            if old_rm:
                old_gw_port_id = old_rm.router['gw_port_id']
            gw_port_id = router['gw_port_id']

            if gw_port_id != old_gw_port_id:
                if old_rm:
                    with IptablesManagerTransaction(old_rm.iptables_manager):
                        self._process_disassociate_metering_label(router)
                        if gw_port_id:
                            self._process_associate_metering_label(router)
                elif gw_port_id:
                    self._process_associate_metering_label(router)

    @log.log
    def remove_router(self, context, router_id):
        if router_id in self.routers:
            del self.routers[router_id]

    def get_external_device_name(self, port_id):
        return (EXTERNAL_DEV_PREFIX + port_id)[:self.driver.DEV_NAME_LEN]

    def _process_metering_label_rules(self, rm, rules, label_chain,
                                      rules_chain):
        im = rm.iptables_manager
        ext_dev = self.get_external_device_name(rm.router['gw_port_id'])
        if not ext_dev:
            return

        for rule in rules:
            remote_ip = rule['remote_ip_prefix']

            if rule['direction'] == 'egress':
                dir_opt = '-o %s -s %s' % (ext_dev, remote_ip)
            else:
                dir_opt = '-i %s -d %s' % (ext_dev, remote_ip)

            if rule['excluded']:
                ipt_rule = '%s -j RETURN' % dir_opt
                im.ipv4['filter'].add_rule(rules_chain, ipt_rule,
                                           wrap=False, top=True)
            else:
                ipt_rule = '%s -j %s' % (dir_opt, label_chain)
                im.ipv4['filter'].add_rule(rules_chain, ipt_rule,
                                           wrap=False, top=False)

    def _process_associate_metering_label(self, router):
        self._update_router(router)
        rm = self.routers.get(router['id'])

        with IptablesManagerTransaction(rm.iptables_manager):
            labels = router.get(constants.METERING_LABEL_KEY, [])
            for label in labels:
                label_id = label['id']

                label_chain = iptables_manager.get_chain_name(WRAP_NAME +
                                                              LABEL + label_id,
                                                              wrap=False)
                rm.iptables_manager.ipv4['filter'].add_chain(label_chain,
                                                             wrap=False)

                rules_chain = iptables_manager.get_chain_name(WRAP_NAME +
                                                              RULE + label_id,
                                                              wrap=False)
                rm.iptables_manager.ipv4['filter'].add_chain(rules_chain,
                                                             wrap=False)
                rm.iptables_manager.ipv4['filter'].add_rule(TOP_CHAIN, '-j ' +
                                                            rules_chain,
                                                            wrap=False)

                rm.iptables_manager.ipv4['filter'].add_rule(label_chain,
                                                            '',
                                                            wrap=False)

                rules = label.get('rules')
                if rules:
                    self._process_metering_label_rules(rm, rules,
                                                       label_chain,
                                                       rules_chain)

                rm.metering_labels[label_id] = label

    def _process_disassociate_metering_label(self, router):
        rm = self.routers.get(router['id'])
        if not rm:
            return

        with IptablesManagerTransaction(rm.iptables_manager):
            labels = router.get(constants.METERING_LABEL_KEY, [])
            for label in labels:
                label_id = label['id']
                if label_id not in rm.metering_labels:
                    continue

                label_chain = iptables_manager.get_chain_name(WRAP_NAME +
                                                              LABEL + label_id,
                                                              wrap=False)
                rules_chain = iptables_manager.get_chain_name(WRAP_NAME +
                                                              RULE + label_id,
                                                              wrap=False)

                rm.iptables_manager.ipv4['filter'].remove_chain(label_chain,
                                                                wrap=False)
                rm.iptables_manager.ipv4['filter'].remove_chain(rules_chain,
                                                                wrap=False)

                del rm.metering_labels[label_id]

    @log.log
    def add_metering_label(self, context, routers):
        for router in routers:
            self._process_associate_metering_label(router)

    @log.log
    def update_metering_label_rules(self, context, routers):
        for router in routers:
            self._update_metering_label_rules(router)

    def _update_metering_label_rules(self, router):
        rm = self.routers.get(router['id'])
        if not rm:
            return

        with IptablesManagerTransaction(rm.iptables_manager):
            labels = router.get(constants.METERING_LABEL_KEY, [])
            for label in labels:
                label_id = label['id']

                label_chain = iptables_manager.get_chain_name(WRAP_NAME +
                                                              LABEL + label_id,
                                                              wrap=False)
                rules_chain = iptables_manager.get_chain_name(WRAP_NAME +
                                                              RULE + label_id,
                                                              wrap=False)
                rm.iptables_manager.ipv4['filter'].empty_chain(rules_chain,
                                                               wrap=False)

                rules = label.get('rules')
                if rules:
                    self._process_metering_label_rules(rm, rules,
                                                       label_chain,
                                                       rules_chain)

    @log.log
    def remove_metering_label(self, context, routers):
        for router in routers:
            self._process_disassociate_metering_label(router)

    @log.log
    def get_traffic_counters(self, context, routers):
        accs = {}
        for router in routers:
            rm = self.routers.get(router['id'])
            if not rm:
                continue

            for label_id, label in rm.metering_labels.items():
                chain = iptables_manager.get_chain_name(WRAP_NAME + LABEL +
                                                        label_id, wrap=False)

                chain_acc = rm.iptables_manager.get_traffic_counters(
                    chain, wrap=False, zero=True)

                if not chain_acc:
                    continue

                acc = accs.get(label_id, {'pkts': 0, 'bytes': 0})

                acc['pkts'] += chain_acc['pkts']
                acc['bytes'] += chain_acc['bytes']

                accs[label_id] = acc

        return accs



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo.config import cfg

from neutron.common import config
from neutron.openstack.common import log as logging


LOG = logging.getLogger(__name__)


ROOT_HELPER_OPTS = [
    cfg.StrOpt('root_helper', default='sudo',
               help=_('Root helper application.')),
]

AGENT_STATE_OPTS = [
    cfg.FloatOpt('report_interval', default=30,
                 help=_('Seconds between nodes reporting state to server; '
                        'should be less than agent_down_time, best if it '
                        'is half or less than agent_down_time.')),
]

INTERFACE_DRIVER_OPTS = [
    cfg.StrOpt('interface_driver',
               help=_("The driver used to manage the virtual interface.")),
]

USE_NAMESPACES_OPTS = [
    cfg.BoolOpt('use_namespaces', default=True,
                help=_("Allow overlapping IP.")),
]


def get_log_args(conf, log_file_name):
    cmd_args = []
    if conf.debug:
        cmd_args.append('--debug')
    if conf.verbose:
        cmd_args.append('--verbose')
    if (conf.log_dir or conf.log_file):
        cmd_args.append('--log-file=%s' % log_file_name)
        log_dir = None
        if conf.log_dir and conf.log_file:
            log_dir = os.path.dirname(
                os.path.join(conf.log_dir, conf.log_file))
        elif conf.log_dir:
            log_dir = conf.log_dir
        elif conf.log_file:
            log_dir = os.path.dirname(conf.log_file)
        if log_dir:
            cmd_args.append('--log-dir=%s' % log_dir)
    else:
        if conf.use_syslog:
            cmd_args.append('--use-syslog')
            if conf.syslog_log_facility:
                cmd_args.append(
                    '--syslog-log-facility=%s' % conf.syslog_log_facility)
    return cmd_args


def register_root_helper(conf):
    # The first call is to ensure backward compatibility
    conf.register_opts(ROOT_HELPER_OPTS)
    conf.register_opts(ROOT_HELPER_OPTS, 'AGENT')


def register_agent_state_opts_helper(conf):
    conf.register_opts(AGENT_STATE_OPTS, 'AGENT')


def register_interface_driver_opts_helper(conf):
    conf.register_opts(INTERFACE_DRIVER_OPTS)


def register_use_namespaces_opts_helper(conf):
    conf.register_opts(USE_NAMESPACES_OPTS)


def get_root_helper(conf):
    root_helper = conf.AGENT.root_helper
    if root_helper != 'sudo':
        return root_helper

    root_helper = conf.root_helper
    if root_helper != 'sudo':
        LOG.deprecated(_('DEFAULT.root_helper is deprecated! Please move '
                         'root_helper configuration to [AGENT] section.'))
        return root_helper

    return 'sudo'


def setup_conf():
    bind_opts = [
        cfg.StrOpt('state_path',
                   default='/var/lib/neutron',
                   help=_('Top-level directory for maintaining dhcp state')),
    ]

    conf = cfg.ConfigOpts()
    conf.register_opts(bind_opts)
    return conf

# add a logging setup method here for convenience
setup_logging = config.setup_logging



#!/usr/bin/env python
# Copyright 2012 Isaku Yamahata <yamahata at private email ne jp>
# Based on openvswitch agent.
#
# Copyright 2011 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# @author: Isaku Yamahata

import httplib
import socket
import sys
import time

import eventlet
from oslo.config import cfg
from ryu.app import client
from ryu.app import conf_switch_key
from ryu.app import rest_nw_id

from neutron.agent.linux import ip_lib
from neutron.agent.linux import ovs_lib
from neutron.agent.linux.ovs_lib import VifPort
from neutron.agent import rpc as agent_rpc
from neutron.agent import securitygroups_rpc as sg_rpc
from neutron.common import config as logging_config
from neutron.common import exceptions as n_exc
from neutron.common import topics
from neutron import context as q_context
from neutron.extensions import securitygroup as ext_sg
from neutron.openstack.common import log
from neutron.openstack.common.rpc import dispatcher
from neutron.plugins.ryu.common import config  # noqa


LOG = log.getLogger(__name__)


# This is copied of nova.flags._get_my_ip()
# Agent shouldn't depend on nova module
def _get_my_ip():
    """Return the actual ip of the local machine.

    This code figures out what source address would be used if some traffic
    were to be sent out to some well known address on the Internet. In this
    case, a Google DNS server is used, but the specific address does not
    matter much.  No traffic is actually sent.
    """
    csock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    csock.connect(('8.8.8.8', 80))
    (addr, _port) = csock.getsockname()
    csock.close()
    return addr


def _get_ip_from_nic(nic):
    ip_wrapper = ip_lib.IPWrapper()
    dev = ip_wrapper.device(nic)
    addrs = dev.addr.list(scope='global')
    for addr in addrs:
        if addr['ip_version'] == 4:
            return addr['cidr'].split('/')[0]


def _get_ip(cfg_ip_str, cfg_interface_str):
    ip = None
    try:
        ip = getattr(cfg.CONF.OVS, cfg_ip_str)
    except (cfg.NoSuchOptError, cfg.NoSuchGroupError):
        pass
    if ip:
        return ip

    iface = None
    try:
        iface = getattr(cfg.CONF.OVS, cfg_interface_str)
    except (cfg.NoSuchOptError, cfg.NoSuchGroupError):
        pass
    if iface:
        ip = _get_ip_from_nic(iface)
        if ip:
            return ip
        LOG.warning(_('Could not get IPv4 address from %(nic)s: %(cfg)s'),
                    {'nic': iface, 'cfg': cfg_interface_str})

    return _get_my_ip()


def _get_tunnel_ip():
    return _get_ip('tunnel_ip', 'tunnel_interface')


def _get_ovsdb_ip():
    return _get_ip('ovsdb_ip', 'ovsdb_interface')


class OVSBridge(ovs_lib.OVSBridge):
    def __init__(self, br_name, root_helper):
        ovs_lib.OVSBridge.__init__(self, br_name, root_helper)
        self.datapath_id = None

    def find_datapath_id(self):
        self.datapath_id = self.get_datapath_id()

    def set_manager(self, target):
        self.run_vsctl(["set-manager", target])

    def get_ofport(self, name):
        return self.db_get_val("Interface", name, "ofport")

    def _get_ports(self, get_port):
        ports = []
        port_names = self.get_port_name_list()
        for name in port_names:
            if self.get_ofport(name) < 0:
                continue
            port = get_port(name)
            if port:
                ports.append(port)

        return ports

    def _get_external_port(self, name):
        # exclude vif ports
        external_ids = self.db_get_map("Interface", name, "external_ids")
        if external_ids:
            return

        # exclude tunnel ports
        options = self.db_get_map("Interface", name, "options")
        if "remote_ip" in options:
            return

        ofport = self.get_ofport(name)
        return VifPort(name, ofport, None, None, self)

    def get_external_ports(self):
        return self._get_ports(self._get_external_port)


class VifPortSet(object):
    def __init__(self, int_br, ryu_rest_client):
        super(VifPortSet, self).__init__()
        self.int_br = int_br
        self.api = ryu_rest_client

    def setup(self):
        for port in self.int_br.get_external_ports():
            LOG.debug(_('External port %s'), port)
            self.api.update_port(rest_nw_id.NW_ID_EXTERNAL,
                                 port.switch.datapath_id, port.ofport)


class RyuPluginApi(agent_rpc.PluginApi,
                   sg_rpc.SecurityGroupServerRpcApiMixin):
    def get_ofp_rest_api_addr(self, context):
        LOG.debug(_("Get Ryu rest API address"))
        return self.call(context,
                         self.make_msg('get_ofp_rest_api'),
                         topic=self.topic)


class RyuSecurityGroupAgent(sg_rpc.SecurityGroupAgentRpcMixin):
    def __init__(self, context, plugin_rpc, root_helper):
        self.context = context
        self.plugin_rpc = plugin_rpc
        self.root_helper = root_helper
        self.init_firewall()


class OVSNeutronOFPRyuAgent(sg_rpc.SecurityGroupAgentRpcCallbackMixin):

    RPC_API_VERSION = '1.1'

    def __init__(self, integ_br, tunnel_ip, ovsdb_ip, ovsdb_port,
                 polling_interval, root_helper):
        super(OVSNeutronOFPRyuAgent, self).__init__()
        self.polling_interval = polling_interval
        self._setup_rpc()
        self.sg_agent = RyuSecurityGroupAgent(self.context,
                                              self.plugin_rpc,
                                              root_helper)
        self._setup_integration_br(root_helper, integ_br, tunnel_ip,
                                   ovsdb_port, ovsdb_ip)

    def _setup_rpc(self):
        self.topic = topics.AGENT
        self.plugin_rpc = RyuPluginApi(topics.PLUGIN)
        self.context = q_context.get_admin_context_without_session()
        self.dispatcher = self._create_rpc_dispatcher()
        consumers = [[topics.PORT, topics.UPDATE],
                     [topics.SECURITY_GROUP, topics.UPDATE]]
        self.connection = agent_rpc.create_consumers(self.dispatcher,
                                                     self.topic,
                                                     consumers)

    def _create_rpc_dispatcher(self):
        return dispatcher.RpcDispatcher([self])

    def _setup_integration_br(self, root_helper, integ_br,
                              tunnel_ip, ovsdb_port, ovsdb_ip):
        self.int_br = OVSBridge(integ_br, root_helper)
        self.int_br.find_datapath_id()

        rest_api_addr = self.plugin_rpc.get_ofp_rest_api_addr(self.context)
        if not rest_api_addr:
            raise n_exc.Invalid(_("Ryu rest API port isn't specified"))
        LOG.debug(_("Going to ofp controller mode %s"), rest_api_addr)

        ryu_rest_client = client.OFPClient(rest_api_addr)

        self.vif_ports = VifPortSet(self.int_br, ryu_rest_client)
        self.vif_ports.setup()

        sc_client = client.SwitchConfClient(rest_api_addr)
        sc_client.set_key(self.int_br.datapath_id,
                          conf_switch_key.OVS_TUNNEL_ADDR, tunnel_ip)

        # Currently Ryu supports only tcp methods. (ssl isn't supported yet)
        self.int_br.set_manager('ptcp:%d' % ovsdb_port)
        sc_client.set_key(self.int_br.datapath_id, conf_switch_key.OVSDB_ADDR,
                          'tcp:%s:%d' % (ovsdb_ip, ovsdb_port))

    def port_update(self, context, **kwargs):
        LOG.debug(_("Port update received"))
        port = kwargs.get('port')
        vif_port = self.int_br.get_vif_port_by_id(port['id'])
        if not vif_port:
            return

        if ext_sg.SECURITYGROUPS in port:
            self.sg_agent.refresh_firewall()

    def _update_ports(self, registered_ports):
        ports = self.int_br.get_vif_port_set()
        if ports == registered_ports:
            return
        added = ports - registered_ports
        removed = registered_ports - ports
        return {'current': ports,
                'added': added,
                'removed': removed}

    def _process_devices_filter(self, port_info):
        if 'added' in port_info:
            self.sg_agent.prepare_devices_filter(port_info['added'])
        if 'removed' in port_info:
            self.sg_agent.remove_devices_filter(port_info['removed'])

    def daemon_loop(self):
        ports = set()

        while True:
            start = time.time()
            try:
                port_info = self._update_ports(ports)
                if port_info:
                    LOG.debug(_("Agent loop has new device"))
                    self._process_devices_filter(port_info)
                    ports = port_info['current']
            except Exception:
                LOG.exception(_("Error in agent event loop"))

            elapsed = max(time.time() - start, 0)
            if (elapsed < self.polling_interval):
                time.sleep(self.polling_interval - elapsed)
            else:
                LOG.debug(_("Loop iteration exceeded interval "
                            "(%(polling_interval)s vs. %(elapsed)s)!"),
                          {'polling_interval': self.polling_interval,
                           'elapsed': elapsed})


def main():
    eventlet.monkey_patch()
    cfg.CONF(project='neutron')

    logging_config.setup_logging(cfg.CONF)

    integ_br = cfg.CONF.OVS.integration_bridge
    polling_interval = cfg.CONF.AGENT.polling_interval
    root_helper = cfg.CONF.AGENT.root_helper

    tunnel_ip = _get_tunnel_ip()
    LOG.debug(_('tunnel_ip %s'), tunnel_ip)
    ovsdb_port = cfg.CONF.OVS.ovsdb_port
    LOG.debug(_('ovsdb_port %s'), ovsdb_port)
    ovsdb_ip = _get_ovsdb_ip()
    LOG.debug(_('ovsdb_ip %s'), ovsdb_ip)
    try:
        agent = OVSNeutronOFPRyuAgent(integ_br, tunnel_ip, ovsdb_ip,
                                      ovsdb_port, polling_interval,
                                      root_helper)
    except httplib.HTTPException as e:
        LOG.error(_("Initialization failed: %s"), e)
        sys.exit(1)

    LOG.info(_("Ryu initialization on the node is done. "
               "Agent initialized successfully, now running..."))
    agent.daemon_loop()
    sys.exit(0)


if __name__ == "__main__":
    main()



# Copyright 2014 Big Switch Networks, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Kevin Benton, kevin.benton@bigswitch.com
#
import httplib
import socket
import ssl

from contextlib import nested
import mock
from oslo.config import cfg

from neutron import context
from neutron.manager import NeutronManager
from neutron.openstack.common import importutils
from neutron.openstack.common import jsonutils
from neutron.plugins.bigswitch.db import consistency_db as cdb
from neutron.plugins.bigswitch import servermanager
from neutron.tests.unit.bigswitch import test_restproxy_plugin as test_rp

HTTPCON = 'httplib.HTTPConnection'
SERVERMANAGER = 'neutron.plugins.bigswitch.servermanager'


class ServerManagerTests(test_rp.BigSwitchProxyPluginV2TestCase):

    def test_no_servers(self):
        cfg.CONF.set_override('servers', [], 'RESTPROXY')
        self.assertRaises(cfg.Error, servermanager.ServerPool)

    def test_malformed_servers(self):
        cfg.CONF.set_override('servers', ['a:b:c'], 'RESTPROXY')
        self.assertRaises(cfg.Error, servermanager.ServerPool)

    def test_sticky_cert_fetch_fail(self):
        pl = NeutronManager.get_plugin()
        pl.servers.ssl = True
        with mock.patch(
            'ssl.get_server_certificate',
            side_effect=Exception('There is no more entropy in the universe')
        ) as sslgetmock:
            self.assertRaises(
                cfg.Error,
                pl.servers._get_combined_cert_for_server,
                *('example.org', 443)
            )
            sslgetmock.assert_has_calls([mock.call(
                ('example.org', 443), ssl_version=ssl.PROTOCOL_TLSv1)])

    def test_consistency_watchdog(self):
        pl = NeutronManager.get_plugin()
        pl.servers.capabilities = []
        self.watch_p.stop()
        with nested(
            mock.patch('eventlet.sleep'),
            mock.patch(
                SERVERMANAGER + '.ServerPool.rest_call',
                side_effect=servermanager.RemoteRestError(
                    reason='Failure to trigger except clause.'
                )
            ),
            mock.patch(
                SERVERMANAGER + '.LOG.exception',
                side_effect=KeyError('Failure to break loop')
            )
        ) as (smock, rmock, lmock):
            # should return immediately without consistency capability
            pl.servers._consistency_watchdog()
            self.assertFalse(smock.called)
            pl.servers.capabilities = ['consistency']
            self.assertRaises(KeyError,
                              pl.servers._consistency_watchdog)
            rmock.assert_called_with('GET', '/health', '', {}, [], False)
            self.assertEqual(1, len(lmock.mock_calls))

    def test_consistency_hash_header(self):
        # mock HTTP class instead of rest_call so we can see headers
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value.getheader.return_value = 'HASHHEADER'
            rv.getresponse.return_value.status = 200
            rv.getresponse.return_value.read.return_value = ''
            with self.network():
                callheaders = rv.request.mock_calls[0][1][3]
                self.assertIn('X-BSN-BVS-HASH-MATCH', callheaders)
                # first call will be empty to indicate no previous state hash
                self.assertEqual(callheaders['X-BSN-BVS-HASH-MATCH'], '')
                # change the header that will be received on delete call
                rv.getresponse.return_value.getheader.return_value = 'HASH2'

            # net delete should have used header received on create
            callheaders = rv.request.mock_calls[1][1][3]
            self.assertEqual(callheaders['X-BSN-BVS-HASH-MATCH'], 'HASHHEADER')

            # create again should now use header received from prev delete
            with self.network():
                callheaders = rv.request.mock_calls[2][1][3]
                self.assertIn('X-BSN-BVS-HASH-MATCH', callheaders)
                self.assertEqual(callheaders['X-BSN-BVS-HASH-MATCH'],
                                 'HASH2')

    def test_consistency_hash_header_no_update_on_bad_response(self):
        # mock HTTP class instead of rest_call so we can see headers
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value.getheader.return_value = 'HASHHEADER'
            rv.getresponse.return_value.status = 200
            rv.getresponse.return_value.read.return_value = ''
            with self.network():
                # change the header that will be received on delete call
                rv.getresponse.return_value.getheader.return_value = 'EVIL'
                rv.getresponse.return_value.status = 'GARBAGE'

            # create again should not use header from delete call
            with self.network():
                callheaders = rv.request.mock_calls[2][1][3]
                self.assertIn('X-BSN-BVS-HASH-MATCH', callheaders)
                self.assertEqual(callheaders['X-BSN-BVS-HASH-MATCH'],
                                 'HASHHEADER')

    def test_file_put_contents(self):
        pl = NeutronManager.get_plugin()
        with mock.patch(SERVERMANAGER + '.open', create=True) as omock:
            pl.servers._file_put_contents('somepath', 'contents')
            omock.assert_has_calls([mock.call('somepath', 'w')])
            omock.return_value.__enter__.return_value.assert_has_calls([
                mock.call.write('contents')
            ])

    def test_combine_certs_to_file(self):
        pl = NeutronManager.get_plugin()
        with mock.patch(SERVERMANAGER + '.open', create=True) as omock:
            omock.return_value.__enter__().read.return_value = 'certdata'
            pl.servers._combine_certs_to_file(['cert1.pem', 'cert2.pem'],
                                              'combined.pem')
            # mock shared between read and write file handles so the calls
            # are mixed together
            omock.assert_has_calls([
                mock.call('combined.pem', 'w'),
                mock.call('cert1.pem', 'r'),
                mock.call('cert2.pem', 'r'),
            ], any_order=True)
            omock.return_value.__enter__.return_value.assert_has_calls([
                mock.call.read(),
                mock.call.write('certdata'),
                mock.call.read(),
                mock.call.write('certdata')
            ])

    def test_req_context_header(self):
        sp = NeutronManager.get_plugin().servers
        ncontext = context.Context('uid', 'tid')
        sp.set_context(ncontext)
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value.getheader.return_value = 'HASHHEADER'
            sp.rest_action('GET', '/')
        callheaders = rv.request.mock_calls[0][1][3]
        self.assertIn(servermanager.REQ_CONTEXT_HEADER, callheaders)
        ctxdct = ncontext.to_dict()
        self.assertEqual(
            ctxdct, jsonutils.loads(
                callheaders[servermanager.REQ_CONTEXT_HEADER]))

    def test_capabilities_retrieval(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value.getresponse.return_value
            rv.getheader.return_value = 'HASHHEADER'

            # each server will get different capabilities
            rv.read.side_effect = ['["a","b","c"]', '["b","c","d"]']
            # pool capabilities is intersection between both
            self.assertEqual(set(['b', 'c']), sp.get_capabilities())
            self.assertEqual(2, rv.read.call_count)

            # the pool should cache after the first call so no more
            # HTTP calls should be made
            rv.read.side_effect = ['["w","x","y"]', '["x","y","z"]']
            self.assertEqual(set(['b', 'c']), sp.get_capabilities())
            self.assertEqual(2, rv.read.call_count)

    def test_capabilities_retrieval_failure(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value.getresponse.return_value
            rv.getheader.return_value = 'HASHHEADER'
            # a failure to parse should result in an empty capability set
            rv.read.return_value = 'XXXXX'
            self.assertEqual([], sp.servers[0].get_capabilities())

            # One broken server should affect all capabilities
            rv.read.side_effect = ['{"a": "b"}', '["b","c","d"]']
            self.assertEqual(set(), sp.get_capabilities())

    def test_reconnect_cached_connection(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value.getheader.return_value = 'HASH'
            sp.servers[0].capabilities = ['keep-alive']
            sp.servers[0].rest_call('GET', '/first')
            # raise an error on re-use to verify reconnect
            # return okay the second time so the reconnect works
            rv.request.side_effect = [httplib.ImproperConnectionState(),
                                      mock.MagicMock()]
            sp.servers[0].rest_call('GET', '/second')
        uris = [c[1][1] for c in rv.request.mock_calls]
        expected = [
            sp.base_uri + '/first',
            sp.base_uri + '/second',
            sp.base_uri + '/second',
        ]
        self.assertEqual(uris, expected)

    def test_no_reconnect_recurse_to_infinity(self):
        # retry uses recursion when a reconnect is necessary
        # this test makes sure it stops after 1 recursive call
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            # hash header must be string instead of mock object
            rv.getresponse.return_value.getheader.return_value = 'HASH'
            sp.servers[0].capabilities = ['keep-alive']
            sp.servers[0].rest_call('GET', '/first')
            # after retrying once, the rest call should raise the
            # exception up
            rv.request.side_effect = httplib.ImproperConnectionState()
            self.assertRaises(httplib.ImproperConnectionState,
                              sp.servers[0].rest_call,
                              *('GET', '/second'))
            # 1 for the first call, 2 for the second with retry
            self.assertEqual(rv.request.call_count, 3)

    def test_socket_error(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            conmock.return_value.request.side_effect = socket.timeout()
            resp = sp.servers[0].rest_call('GET', '/')
            self.assertEqual(resp, (0, None, None, None))

    def test_retry_on_unavailable(self):
        pl = NeutronManager.get_plugin()
        with nested(
            mock.patch(SERVERMANAGER + '.ServerProxy.rest_call',
                       return_value=(httplib.SERVICE_UNAVAILABLE, 0, 0, 0)),
            mock.patch(SERVERMANAGER + '.time.sleep')
        ) as (srestmock, tmock):
            # making a call should trigger retries with sleeps in between
            pl.servers.rest_call('GET', '/', '', None, [])
            rest_call = [mock.call('GET', '/', '', None, False, reconnect=True,
                                   hash_handler=mock.ANY)]
            rest_call_count = (
                servermanager.HTTP_SERVICE_UNAVAILABLE_RETRY_COUNT + 1)
            srestmock.assert_has_calls(rest_call * rest_call_count)
            sleep_call = [mock.call(
                servermanager.HTTP_SERVICE_UNAVAILABLE_RETRY_INTERVAL)]
            # should sleep 1 less time than the number of calls
            sleep_call_count = rest_call_count - 1
            tmock.assert_has_calls(sleep_call * sleep_call_count)

    def test_delete_failure_sets_bad_hash(self):
        pl = NeutronManager.get_plugin()
        hash_handler = cdb.HashHandler()
        with mock.patch(
            SERVERMANAGER + '.ServerProxy.rest_call',
            return_value=(httplib.INTERNAL_SERVER_ERROR, 0, 0, 0)
        ):
            # a failed delete call should put a bad hash in the DB
            pl.servers.rest_call('DELETE', '/', '', None, [])
            self.assertEqual('INCONSISTENT,INCONSISTENT',
                             hash_handler.read_for_update())


class TestSockets(test_rp.BigSwitchProxyPluginV2TestCase):

    def setUp(self):
        super(TestSockets, self).setUp()
        # http patch must not be running or it will mangle the servermanager
        # import where the https connection classes are defined
        self.httpPatch.stop()
        self.sm = importutils.import_module(SERVERMANAGER)

    def test_socket_create_attempt(self):
        # exercise the socket creation to make sure it works on both python
        # versions
        con = self.sm.HTTPSConnectionWithValidation('127.0.0.1', 0, timeout=1)
        # if httpcon was created, a connect attempt should raise a socket error
        self.assertRaises(socket.error, con.connect)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import types

import fixtures

from oslo.config import cfg

from neutron.common import config
from neutron.manager import NeutronManager
from neutron.manager import validate_post_plugin_load
from neutron.manager import validate_pre_plugin_load
from neutron.openstack.common import log as logging
from neutron.plugins.common import constants
from neutron.tests import base
from neutron.tests.unit import dummy_plugin


LOG = logging.getLogger(__name__)
DB_PLUGIN_KLASS = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'
ROOTDIR = os.path.dirname(os.path.dirname(__file__))
ETCDIR = os.path.join(ROOTDIR, 'etc')


def etcdir(*p):
    return os.path.join(ETCDIR, *p)


class MultiServiceCorePlugin(object):
    supported_extension_aliases = ['lbaas', 'dummy']


class CorePluginWithAgentNotifiers(object):
    agent_notifiers = {'l3': 'l3_agent_notifier',
                       'dhcp': 'dhcp_agent_notifier'}


class NeutronManagerTestCase(base.BaseTestCase):

    def setUp(self):
        super(NeutronManagerTestCase, self).setUp()
        args = ['--config-file', etcdir('neutron.conf.test')]
        # If test_config specifies some config-file, use it, as well
        config.parse(args=args)
        self.setup_coreplugin()
        self.useFixture(
            fixtures.MonkeyPatch('neutron.manager.NeutronManager._instance'))

    def test_service_plugin_is_loaded(self):
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin"])
        mgr = NeutronManager.get_instance()
        plugin = mgr.get_service_plugins()[constants.DUMMY]

        self.assertTrue(
            isinstance(plugin,
                       (dummy_plugin.DummyServicePlugin, types.ClassType)),
            "loaded plugin should be of type neutronDummyPlugin")

    def test_service_plugin_by_name_is_loaded(self):
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        cfg.CONF.set_override("service_plugins", ["dummy"])
        mgr = NeutronManager.get_instance()
        plugin = mgr.get_service_plugins()[constants.DUMMY]

        self.assertTrue(
            isinstance(plugin,
                       (dummy_plugin.DummyServicePlugin, types.ClassType)),
            "loaded plugin should be of type neutronDummyPlugin")

    def test_multiple_plugins_specified_for_service_type(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin",
                               "neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin"])
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        self.assertRaises(ValueError, NeutronManager.get_instance)

    def test_multiple_plugins_by_name_specified_for_service_type(self):
        cfg.CONF.set_override("service_plugins", ["dummy", "dummy"])
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        self.assertRaises(ValueError, NeutronManager.get_instance)

    def test_multiple_plugins_mixed_specified_for_service_type(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin", "dummy"])
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        self.assertRaises(ValueError, NeutronManager.get_instance)

    def test_service_plugin_conflicts_with_core_plugin(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin"])
        cfg.CONF.set_override("core_plugin",
                              "neutron.tests.unit.test_neutron_manager."
                              "MultiServiceCorePlugin")
        self.assertRaises(ValueError, NeutronManager.get_instance)

    def test_core_plugin_supports_services(self):
        cfg.CONF.set_override("core_plugin",
                              "neutron.tests.unit.test_neutron_manager."
                              "MultiServiceCorePlugin")
        mgr = NeutronManager.get_instance()
        svc_plugins = mgr.get_service_plugins()
        self.assertEqual(3, len(svc_plugins))
        self.assertIn(constants.CORE, svc_plugins.keys())
        self.assertIn(constants.LOADBALANCER, svc_plugins.keys())
        self.assertIn(constants.DUMMY, svc_plugins.keys())

    def test_post_plugin_validation(self):
        cfg.CONF.import_opt('dhcp_agents_per_network',
                            'neutron.db.agentschedulers_db')

        self.assertIsNone(validate_post_plugin_load())
        cfg.CONF.set_override('dhcp_agents_per_network', 2)
        self.assertIsNone(validate_post_plugin_load())
        cfg.CONF.set_override('dhcp_agents_per_network', 0)
        self.assertIsNotNone(validate_post_plugin_load())
        cfg.CONF.set_override('dhcp_agents_per_network', -1)
        self.assertIsNotNone(validate_post_plugin_load())

    def test_pre_plugin_validation(self):
        self.assertIsNotNone(validate_pre_plugin_load())
        cfg.CONF.set_override('core_plugin', 'dummy.plugin')
        self.assertIsNone(validate_pre_plugin_load())

    def test_manager_gathers_agent_notifiers_from_service_plugins(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin"])
        cfg.CONF.set_override("core_plugin",
                              "neutron.tests.unit.test_neutron_manager."
                              "CorePluginWithAgentNotifiers")
        expected = {'l3': 'l3_agent_notifier',
                    'dhcp': 'dhcp_agent_notifier',
                    'dummy': 'dummy_agent_notifier'}
        core_plugin = NeutronManager.get_plugin()
        self.assertEqual(expected, core_plugin.agent_notifiers)
